import os
from typing import Any, Dict, List, Optional

import bionty as bt
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
import torch
from anndata import AnnData, concat
from scdataloader import Collator, Preprocessor
from scdataloader.data import SimpleAnnDataset
from scdataloader.utils import get_descendants, random_str
from scib_metrics.benchmark import Benchmarker
from scipy.stats import spearmanr
from simpler_flash import FlashTransformer
from sklearn.metrics import f1_score
from torch.utils.data import DataLoader
from tqdm import tqdm

FILE_LOC = os.path.dirname(os.path.realpath(__file__))


class Embedder:
    def __init__(
        self,
        batch_size: int = 64,
        num_workers: int = 8,
        how: str = "random expr",
        max_len: int = 2000,
        doclass: bool = True,
        pred_embedding: List[str] = [
            "all",
        ],
        doplot: bool = True,
        keep_all_labels_pred: bool = False,
        genelist: Optional[List[str]] = None,
        save_every: int = 40_000,
        unknown_label: str = "unknown",
        use_knn: bool = True,
    ):
        """
        Embedder a class to embed and annotate cells using a model

        Args:
            batch_size (int, optional): The size of the batches to be used in the DataLoader. Defaults to 64.
            num_workers (int, optional): The number of worker processes to use for data loading. Defaults to 8.
            how (str, optional): The method to be used for selecting valid genes. Defaults to "random expr".
                - "random expr": random expression
                - "most var": highly variable genes in the dataset
                - "some": specific genes (from genelist)
                - "most expr": most expressed genes in the cell
            max_len (int, optional): The maximum length of the gene sequence given to the model. Defaults to 1000.
            doclass (bool, optional): Whether to perform classification. Defaults to True.
            pred_embedding (List[str], optional): The list of labels to be used for plotting embeddings. Defaults to [ "cell_type_ontology_term_id", "disease_ontology_term_id", "self_reported_ethnicity_ontology_term_id", "sex_ontology_term_id", ].
            doplot (bool, optional): Whether to generate plots. Defaults to True.
            keep_all_labels_pred (bool, optional): Whether to keep all class predictions. Defaults to False, will only keep the most likely class.
            genelist (List[str], optional): The list of genes to be used for embedding. Defaults to []: In this case, "how" needs to be "most var" or "random expr".
            save_every (int, optional): The number of cells to save at a time. Defaults to 100_000.
                This is important to avoid memory issues.
            unknown_label (str, optional): The label to be used for unknown cell types. Defaults to "unknown".
            use_knn (bool, optional): Whether to use k-nearest neighbors information. Defaults to True.
        """
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.how = how
        self.max_len = max_len
        self.pred_embedding = pred_embedding
        self.keep_all_labels_pred = keep_all_labels_pred
        self.doplot = doplot
        self.doclass = doclass
        self.genelist = genelist if genelist is not None else []
        self.save_every = save_every
        self.pred = None
        self.unknown_label = unknown_label
        self.use_knn = use_knn

    def __call__(self, model: torch.nn.Module, adata: AnnData) -> tuple[AnnData, dict]:
        """
        __call__ function to call the embedding

        Args:
            model (torch.nn.Module): The scPRINT model to be used for embedding and annotation.
            adata (AnnData): The annotated data matrix of shape n_obs x n_vars. Rows correspond to cells and columns to genes.

        Raises:
            ValueError: If the model does not have a logger attribute.
            ValueError: If the model does not have a global_step attribute.

        Returns:
            AnnData: The annotated data matrix with embedded cell representations.
            dict: classification metrics results when some ground truth information was available in the anndata.
        """
        # one of "all" "sample" "none"
        model.predict_mode = "none"
        self.pred = None
        prevkeep = model.keep_all_labels_pred
        model.keep_all_labels_pred = self.keep_all_labels_pred
        # Add at least the organism you are working with
        if self.how == "most var":
            sc.pp.highly_variable_genes(
                adata, flavor="seurat_v3", n_top_genes=self.max_len
            )
            self.genelist = adata.var.index[adata.var.highly_variable]
        adataset = SimpleAnnDataset(
            adata,
            obs_to_output=["organism_ontology_term_id"],
            get_knn_cells=model.expr_emb_style == "metacell" and self.use_knn,
        )
        col = Collator(
            organisms=model.organisms,
            valid_genes=model.genes,
            how=self.how if self.how != "most var" else "some",
            max_len=self.max_len,
            add_zero_genes=0,
            genelist=self.genelist if self.how in ["most var", "some"] else [],
            n_bins=model.n_input_bins if model.expr_emb_style == "binned" else 0,
        )
        dataloader = DataLoader(
            adataset,
            collate_fn=col,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=False,
        )
        model.eval()
        model.on_predict_epoch_start()
        device = model.device.type
        prevplot = model.doplot
        model.pred_log_adata = True
        model.doplot = self.doplot and not self.keep_all_labels_pred
        model.save_expr = False
        rand = random_str()
        dtype = (
            torch.float16
            if isinstance(model.transformer, FlashTransformer)
            else model.dtype
        )
        with (
            torch.no_grad(),
            torch.autocast(device_type=device, dtype=dtype),
        ):
            for batch in tqdm(dataloader):
                gene_pos, expression, depth = (
                    batch["genes"].to(device),
                    batch["x"].to(device),
                    batch["depth"].to(device),
                )
                pred = model._predict(
                    gene_pos,
                    expression,
                    depth,
                    knn_cells=(
                        batch["knn_cells"].to(device)
                        if model.expr_emb_style == "metacell" and self.use_knn
                        else None
                    ),
                    knn_cells_info=(
                        batch["knn_cells_info"].to(device)
                        if model.expr_emb_style == "metacell" and self.use_knn
                        else None
                    ),
                    pred_embedding=self.pred_embedding,
                    max_size_in_mem=self.save_every,
                    name="embed_" + rand + "_",
                )
                torch.cuda.empty_cache()
                if self.keep_all_labels_pred:
                    if pred is not None:
                        self.pred = (
                            pred if self.pred is None else torch.cat([self.pred, pred])
                        )
        model.log_adata(name="embed_" + rand + "_" + str(model.counter))

        model.pos = None
        model.expr_pred = None
        model.embs = None
        if self.keep_all_labels_pred:
            self.pred = (
                model.pred if self.pred is None else torch.cat([self.pred, model.pred])
            )
        model.pred = None
        model.save_expr = True
        try:
            mdir = (
                model.logger.save_dir if model.logger.save_dir is not None else "data"
            )
        except:
            mdir = "data"
        pred_adata = []
        del adataset, dataloader
        for i in range(model.counter + 1):
            file = (
                mdir
                + "/step_"
                + str(model.global_step)
                + "_"
                + model.name
                + "_embed_"
                + rand
                + "_"
                + str(i)
                + "_"
                + str(model.global_rank)
                + ".h5ad"
            )
            pred_adata.append(sc.read_h5ad(file))
            os.remove(file)
        pred_adata = concat(pred_adata)
        pred_adata.obs.index = adata.obs.index

        try:
            adata.obsm["X_scprint_umap"] = pred_adata.obsm["X_umap"]
            del pred_adata.obsm["X_umap"]
        except:
            print("too few cells to embed into a umap")
        try:
            adata.obs["scprint_leiden"] = pred_adata.obs["scprint_leiden"]
        except:
            print("too few cells to compute a clustering")

        if self.pred_embedding == ["all"]:
            pred_embedding = ["other"] + model.classes
        else:
            pred_embedding = self.pred_embedding
        if len(pred_embedding) == 1:
            adata.obsm["scprint_emb"] = pred_adata.obsm[
                "scprint_emb_" + pred_embedding[0]
            ].astype(np.float32)

        else:
            adata.obsm["scprint_emb"] = np.zeros(
                pred_adata.obsm["scprint_emb_" + pred_embedding[0]].shape,
                dtype=np.float32,
            )
            i = 0
            for k, v in pred_adata.obsm.items():
                adata.obsm[k] = v.astype(np.float32)
                if model.compressor is not None:
                    if i == 0:
                        adata.obsm["scprint_emb"] = v.astype(np.float32)
                    else:
                        adata.obsm["scprint_emb"] = np.hstack(
                            [adata.obsm["scprint_emb"], v.astype(np.float32)]
                        )
                else:
                    adata.obsm["scprint_emb"] += v.astype(np.float32)
                i += 1
            if model.compressor is None:
                adata.obsm["scprint_emb"] = adata.obsm["scprint_emb"] / i

        for key, value in pred_adata.uns.items():
            adata.uns[key] = value

        pred_adata.obs.index = adata.obs.index
        model.keep_all_labels_pred = prevkeep
        model.doplot = prevplot
        adata.obs = pd.concat([adata.obs, pred_adata.obs], axis=1)
        del pred_adata
        if self.keep_all_labels_pred:
            allclspred = self.pred.to(device="cpu").numpy()
            columns = []
            for cl in model.classes:
                n = model.label_counts[cl]
                columns += [model.label_decoders[cl][i] for i in range(n)]
            allclspred = pd.DataFrame(
                allclspred, columns=columns, index=adata.obs.index
            )
            adata.obs = pd.concat([adata.obs, allclspred], axis=1)

        metrics = {}
        if self.doclass and not self.keep_all_labels_pred:
            for cl in model.classes:
                res = []
                if cl not in adata.obs.columns:
                    continue
                class_topred = model.label_decoders[cl].values()

                if cl in model.labels_hierarchy:
                    # class_groupings = {
                    #    k: [
                    #        i.ontology_id
                    #        for i in bt.CellType.filter(k).first().children.all()
                    #    ]
                    #    for k in set(adata.obs[cl].unique()) - set(class_topred)
                    # }
                    cur_labels_hierarchy = {
                        model.label_decoders[cl][k]: [
                            model.label_decoders[cl][i] for i in v
                        ]
                        for k, v in model.labels_hierarchy[cl].items()
                    }
                else:
                    cur_labels_hierarchy = {}

                for pred, true in adata.obs[["pred_" + cl, cl]].values:
                    if pred == true:
                        res.append(True)
                        continue
                    if len(cur_labels_hierarchy) > 0:
                        if true in cur_labels_hierarchy:
                            res.append(pred in cur_labels_hierarchy[true])
                            continue
                        elif true != self.unknown_label:
                            res.append(False)
                        elif true not in class_topred:
                            print(f"true label {true} not in available classes")
                            return adata, metrics
                    elif true not in class_topred:
                        print(f"true label {true} not in available classes")
                        return adata, metrics
                    elif true != self.unknown_label:
                        res.append(False)
                    # else true is unknown
                    # else we pass
                if len(res) == 0:
                    # true was always unknown
                    res = [1]
                if self.doplot:
                    print("    ", cl)
                    print("     accuracy:", sum(res) / len(res))
                    print(" ")
                metrics.update({cl + "_accuracy": sum(res) / len(res)})
        self.pred = None
        return adata, metrics


def compute_corr(
    out: np.ndarray,
    to: np.ndarray,
    doplot: bool = True,
    compute_mean_regress: bool = False,
    plot_corr_size: int = 64,
) -> dict:
    """
    Compute the correlation between the output and target matrices.

    Args:
        out (np.ndarray): The output matrix.
        to (np.ndarray): The target matrix.
        doplot (bool, optional): Whether to generate a plot of the correlation coefficients. Defaults to True.
        compute_mean_regress (bool, optional): Whether to compute mean regression. Defaults to False.
        plot_corr_size (int, optional): The size of the plot for correlation. Defaults to 64.

    Returns:
        dict: A dictionary containing the computed metrics.
    """
    metrics = {}
    corr_coef, p_value = spearmanr(
        out,
        to.T,
    )
    corr_coef[p_value > 0.05] = 0
    # corr_coef[]
    # only on non zero values,
    # compare a1-b1 corr with a1-b(n) corr. should be higher

    # Plot correlation coefficient
    val = plot_corr_size + 2 if compute_mean_regress else plot_corr_size
    metrics.update(
        {"recons_corr": np.mean(corr_coef[val:, :plot_corr_size].diagonal())}
    )
    if compute_mean_regress:
        metrics.update(
            {
                "mean_regress": np.mean(
                    corr_coef[
                        plot_corr_size : plot_corr_size + 2,
                        :plot_corr_size,
                    ].flatten()
                )
            }
        )
    if doplot:
        plt.figure(figsize=(10, 5))
        plt.imshow(corr_coef, cmap="coolwarm", interpolation="none", vmin=-1, vmax=1)
        plt.colorbar()
        plt.title('Correlation Coefficient of expr and i["x"]')
        plt.show()
    return metrics


def default_benchmark(
    model: torch.nn.Module,
    folder_dir: str = FILE_LOC + "/../../data/",
    dataset: str = FILE_LOC + "/../../data/gNNpgpo6gATjuxTE7CCp.h5ad",
    do_class: bool = True,
    coarse: bool = False,
) -> dict:
    """
    Run the default benchmark for embedding and annotation using the scPRINT model.

    Args:
        model (torch.nn.Module): The scPRINT model to be used for embedding and annotation.
        folder_dir (str, optional): The directory containing data files.
        dataset (str, optional): The dataset to use for benchmarking. Can be a path or URL.
        do_class (bool, optional): Whether to perform classification. Defaults to True.
        coarse (bool, optional): Whether to use coarse cell type annotations. Defaults to False.

    Returns:
        dict: A dictionary containing the benchmark metrics.
    """
    if dataset.startswith("https://"):
        adata = sc.read(
            folder_dir
            + dataset.split("/")[-1]
            + (".h5ad" if not dataset.endswith(".h5ad") else ""),
            backup_url=dataset,
        )
    else:
        adata = sc.read_h5ad(dataset)
    if adata.shape[0] > 100_000:
        adata = adata[
            adata.obs_names[np.random.choice(adata.shape[0], 100_000, replace=False)]
        ]
    max_len = 4000 if adata.X.sum(1).mean() < 50_000 else 8000
    batch_size = 64 if adata.X.sum(1).mean() < 50_000 else 32
    log_every = 10_000
    if dataset.split("/")[-1] in ["24539942", "24539828"]:  # lung and pancreas
        adata.obs["organism_ontology_term_id"] = "NCBITaxon:9606"
        use_layer = "counts"
        is_symbol = True
        batch_key = "tech" if dataset.split("/")[-1] == "24539828" else "batch"
        label_key = "celltype" if dataset.split("/")[-1] == "24539828" else "cell_type"
        adata.obs["cell_type_ontology_term_id"] = adata.obs[label_key].replace(
            COARSE if coarse else FINE
        )
        adata.obs["assay_ontology_term_id"] = adata.obs[batch_key].replace(
            COARSE if coarse else FINE
        )
    else:
        use_layer = None
        is_symbol = False
        batch_key = (
            "batch"
            if dataset.split("/")[-1] == "661d5ec2-ca57-413c-8374-f49b0054ddba.h5ad"
            else "assay_ontology_term_id"
        )
        label_key = "cell_type_ontology_term_id"
    preprocessor = Preprocessor(
        use_layer=use_layer,
        is_symbol=is_symbol,
        force_preprocess=True,
        skip_validate=True,
        do_postp=model.expr_emb_style == "metacell",
        drop_non_primary=False,
    )
    adata = preprocessor(adata.copy())
    if model.expr_emb_style == "metacell":
        sc.pp.neighbors(adata, use_rep="X_pca")
    embedder = Embedder(
        pred_embedding=(
            model.pred_embedding if model.pred_embedding is not None else ["all"]
        ),
        doclass=do_class,
        max_len=max_len,
        doplot=False,
        keep_all_labels_pred=False,
        save_every=log_every,
        batch_size=batch_size,
        how="random expr",
    )
    adata, metrics = embedder(model, adata)

    bm = Benchmarker(
        adata,
        batch_key=batch_key,
        label_key=label_key,
        embedding_obsm_keys=["scprint_emb"],
    )
    bm.benchmark()
    metrics.update(
        {"scib": bm.get_results(min_max_scale=False).T.to_dict()["scprint_emb"]}
    )
    if model.class_scale > 0:
        metrics["classif"] = compute_classification(
            adata, model.classes, model.label_decoders, model.labels_hierarchy
        )
    return metrics


def compute_classification(
    adata: AnnData,
    classes: List[str],
    label_decoders: Dict[str, Any],
    labels_hierarchy: Dict[str, Any],
    metric_type: List[str] = ["macro", "micro", "weighted"],
    use_unknown: bool = False,
) -> Dict[str, Dict[str, float]]:
    """
    Compute classification metrics for the given annotated data.

    Args:
        adata (AnnData): The annotated data matrix of shape n_obs x n_vars. Rows correspond to cells and columns to genes.
        classes (List[str]): List of class labels to be used for classification.
        label_decoders (Dict[str, Any]): Dictionary of label decoders for each class.
        labels_hierarchy (Dict[str, Any]): Dictionary representing the hierarchy of labels.
        metric_type (List[str], optional): List of metric types to compute. Defaults to ["macro", "micro", "weighted"].

    Returns:
        Dict[str, Dict[str, float]]: A dictionary containing classification metrics for each class.
    """
    metrics = {}
    for clss in classes:
        res = []
        if clss not in adata.obs.columns:
            print("not in columns")
            continue
        labels_topred = label_decoders[clss].values()
        if clss in labels_hierarchy:
            parentdf = (
                bt.CellType.filter()
                .to_dataframe(
                    include=["parents__ontology_id", "ontology_id"], limit=None
                )
                .set_index("ontology_id")[["parents__ontology_id"]]
            )
            parentdf.parents__ontology_id = parentdf.parents__ontology_id.astype(str)
            class_groupings = {
                k: get_descendants(k, parentdf) for k in set(adata.obs[clss].unique())
            }
        tokeep = np.array([True] * adata.shape[0])
        for i, (pred, true) in enumerate(adata.obs[["pred_" + clss, clss]].values):
            if pred == true:
                res.append(true)
                continue
            if true == "unknown":
                tokeep[i] = False
            if clss in labels_hierarchy:
                if true in class_groupings:
                    if pred == "unknown" and not use_unknown:
                        tokeep[i] = False
                    res.append(true if pred in class_groupings[true] else "")
                    continue
                elif true not in labels_topred:
                    raise ValueError(f"true label {true} not in available classes")
            elif true not in labels_topred:
                raise ValueError(f"true label {true} not in available classes")
            res.append("")
        metrics[clss] = {}
        metrics[clss]["accuracy"] = np.mean(
            np.array(res)[tokeep] == adata.obs[clss].values[tokeep]
        )
        for x in metric_type:
            metrics[clss][x] = f1_score(
                np.array(res)[tokeep], adata.obs[clss].values[tokeep], average=x
            )
    return metrics


def get_top_labels(model, adata, k=3):
    s = list(model.label_counts.values())
    out = []
    count = 0
    for l, label in enumerate(model.label_counts.keys()):
        if -sum(s) + count + s[l] == 0:
            ct = adata.obs.iloc[:, -sum(s) + count :]
        else:
            ct = adata.obs.iloc[:, -sum(s) + count : -sum(s) + s[l] + count]
        if s[l] < 3:
            # append only the label corresponding to the maximum score (arg-max)
            best_labels = [
                model.label_decoders[label][idx]
                for idx in np.argmax(ct.to_numpy(), axis=1)
            ]
            out.append(
                pd.DataFrame(
                    {label: best_labels}, index=adata.obs_names, columns=[label]
                )
            )
        else:
            res = []
            for j, m in enumerate(np.argsort(ct)[:, ::-1]):
                certainty = ct.iloc[j, m[0]]
                best = model.label_decoders[label][m[0]]
                other = []
                for i in range(1, k):
                    if ct.iloc[j, m[i]] > certainty - 0.15:
                        other.append(model.label_decoders[label][m[i]])
                    else:
                        other.append(None)
                res.append([best] + [certainty] + other)
            out.append(
                pd.DataFrame(
                    res,
                    columns=[label, label + "_certainty"]
                    + [label + "_choice" + str(i) for i in range(1, k)],
                    index=adata.obs_names,
                )
            )
        count += s[l]  # keep the offset consistent with the else branch
    out = pd.concat(out, axis=1)
    return out


def find_coarser_labels(out, model):
    relabel = {}
    for label in model.classes:
        if label in model.labels_hierarchy.keys():
            decoder = {k: v for k, v in model.label_decoders[label].items()}
            ct_hier = {
                decoder[k]: [decoder[u] for u in v]
                for k, v in model.labels_hierarchy[label].items()
            }
            relabel[label] = {}
            for i, r in enumerate(
                out[~out[label + "_choice1"].isna()][
                    [label, label + "_choice1", label + "_choice2"]
                ].values
            ):
                prev_v = 100_000
                res = None
                if r[1] is not None:
                    for k, v in ct_hier.items():
                        if r[2] is not None:
                            if (
                                r[0] in v
                                and r[1] in v
                                and r[2] in v
                                and k != "CL:0000000"
                                and k != "None"
                            ):
                                if len(v) < prev_v:
                                    res = k
                                    prev_v = len(v)
                        else:
                            if (
                                r[0] in v
                                and r[1] in v
                                and k != "CL:0000000"
                                and k != "None"
                            ):
                                if len(v) < prev_v:
                                    res = k
                                    prev_v = len(v)
                    if res is not None:
                        relabel[label].update({i: res})
    return relabel


def display_confusion_matrix(
    nadata, pred="conv_pred_cell_type_ontology_term_id", true="cell_type"
):
    """
    Display the confusion matrix for true vs predicted cell types.

    Args:
        nadata (AnnData): Annotated data object containing predictions and ground truth.
        pred (str): Column name for predictions. Defaults to "conv_pred_cell_type_ontology_term_id".
        true (str): Column name for ground truth. Defaults to "cell_type".
    """
    counts = None
    for k, v in nadata.obs[true].value_counts().items():
        name = k + " - " + str(v)
        if counts is None:
            counts = pd.DataFrame(
                nadata.obs.loc[
                    nadata.obs[true] == k,
                    pred,
                ].value_counts()
            ).rename(columns={"count": name})
        else:
            counts = pd.concat(
                [
                    counts,
                    pd.DataFrame(
                        nadata.obs.loc[
                            nadata.obs[true] == k,
                            pred,
                        ].value_counts(),
                    ).rename(columns={"count": name}),
                ],
                axis=1,
            )
    counts = counts.T
    # Fill NaN values with 0 for visualization
    counts_filled = counts.fillna(0)

    # Create the heatmap
    plt.figure(figsize=(12, 10))

    # Convert to percentages (row-wise normalization)
    counts_percentage = counts_filled.div(counts_filled.sum(axis=1), axis=0) * 100
    counts_percentage = counts_percentage.iloc[:, counts_percentage.values.max(0) > 5]

    ax = sns.heatmap(
        counts_percentage,
        cmap="Blues",
        cbar_kws={"label": "Percentage (%)"},
        linewidths=0.5,
        square=True,
    )
    # place the x-label on top
    ax.xaxis.set_label_position("top")
    ax.xaxis.tick_top()

    plt.title(
        "Confusion Matrix: " + true + " vs " + pred + " (Percentage)",
        fontsize=16,
        pad=20,
    )
    ax.set_xlabel(pred, fontsize=12)
    ax.set_ylabel(true + " (with counts)", fontsize=12)
    ax.set_xticklabels(ax.get_xticklabels(), rotation=45, ha="left", fontsize=12)
    ax.set_yticklabels(ax.get_yticklabels(), rotation=0, fontsize=14)
    plt.tight_layout()
    plt.show()


FINE = {
    "gamma": "CL:0002275",
    "beta": "CL:0000169",  # "CL:0008024"
    "epsilon": "CL:0005019",  # "CL:0008024"
    "acinar": "CL:0000622",
    "delta": "CL:0000173",  # "CL:0008024"
    "schwann": "CL:0002573",  # "CL:0000125"
    "activated_stellate": "CL:0000057",
    "alpha": "CL:0000171",  # "CL:0008024"
    "mast": "CL:0000097",
    "Mast cell": "CL:0000097",
    "quiescent_stellate": "CL:0000057",
    "t_cell": "CL:0000084",
    "endothelial": "CL:0000115",
    "Endothelium": "CL:0000115",
    "ductal": "CL:0002079",  # CL:0000068
    "macrophage": "CL:0000235",
    "Macrophage": "CL:0000235",
    "B cell": "CL:0000236",
    "Type 2": "CL:0002063",
    "Type 1": "CL:0002062",
    "Ciliated": "CL:4030034",  # respiratory ciliated
    "Dendritic cell": "CL:0000451",  # leukocyte
    "Ionocytes": "CL:0005006",
    "Basal 1": "CL:0000646",  # epithelial
    "Basal 2": "CL:0000646",
    "Secretory": "CL:0000151",
    "Neutrophil_CD14_high": "CL:0000775",
    "Neutrophils_IL1R2": "CL:0000775",
    "Lymphatic": "CL:0002138",
    "Fibroblast": "CL:0000057",
    "T/NK cell": "CL:0000814",
    "inDrop1": "EFO:0008780",
    "inDrop3": "EFO:0008780",
    "inDrop4": "EFO:0008780",
    "inDrop2": "EFO:0008780",
    "fluidigmc1": "EFO:0010058",  # fluidigm c1
    "smarter": "EFO:0010058",  # fluidigm c1
    "celseq2": "EFO:0010010",
    "smartseq2": "EFO:0008931",
    "celseq": "EFO:0008679",
}
COARSE = {
    "beta": "CL:0008024",  # endocrine
    "epsilon": "CL:0008024",
    "delta": "CL:0008024",
    "alpha": "CL:0008024",
    "gamma": "CL:0008024",
    "acinar": "CL:0000150",  # epithelial (gland)
    "ductal": "CL:0000068",  # epithelial (duct)
    "schwann": "CL:0000125",  # glial
    "endothelial": "CL:0000115",
    "Endothelium": "CL:0000115",
    "Lymphatic": "CL:0000115",
    "macrophage": "CL:0000235",  # myeloid leukocyte (not)
    "Macrophage": "CL:0000235",  # myeloid leukocyte
    "mast": "CL:0000097",  # myeloid leukocyte (not)
    "Mast cell": "CL:0000097",  # myeloid leukocyte
    "Neutrophil_CD14_high": "CL:0000775",  # myeloid leukocyte
    "Neutrophils_IL1R2": "CL:0000775",  # myeloid leukocyte
    "t_cell": "CL:0000084",  # leukocyte, lymphocyte (not)
    "T/NK cell": "CL:0000084",  # leukocyte, lymphocyte (not)
    "B cell": "CL:0000236",  # leukocyte, lymphocyte (not)
    "Dendritic cell": "CL:0000451",  # leukocyte, lymphocyte
    "activated_stellate": "CL:0000057",  # fibroblast (not)
    "quiescent_stellate": "CL:0000057",  # fibroblast (not)
    "Fibroblast": "CL:0000057",
    "Type 2": "CL:0000066",  # epithelial
    "Type 1": "CL:0000066",
    "Ionocytes": "CL:0000066",  # epithelial
    "Basal 1": "CL:0000066",  # epithelial
    "Basal 2": "CL:0000066",
    "Ciliated": "CL:0000064",  # ciliated
    "Secretory": "CL:0000151",
    "inDrop1": "EFO:0008780",
    "inDrop3": "EFO:0008780",
    "inDrop4": "EFO:0008780",
    "inDrop2": "EFO:0008780",
    "fluidigmc1": "EFO:0010058",  # fluidigm c1
    "smarter": "EFO:0010058",  # fluidigm c1
    "celseq2": "EFO:0010010",
    "smartseq2": "EFO:0008931",
    "celseq": "EFO:0008679",
}



import json
import os
import random
import subprocess
from collections import OrderedDict
from itertools import repeat
from pathlib import Path
from typing import List

import bionty as bt
import colorcet as cc
import lamindb as ln
import numpy as np
import pandas as pd
import torch
from bokeh.io import export_svg, show
from bokeh.layouts import column
from bokeh.models import ColumnDataSource, CustomJS, HoverTool, TextInput
from bokeh.models.annotations import LabelSet
from bokeh.palettes import Category10, Category20
from bokeh.plotting import figure, save
from IPython import get_ipython

# What pops up on hover?
from matplotlib import pyplot as plt


def run_command(command: str, **kwargs) -> int:
    """
    run_command runs a command in the shell and prints the output.

    Args:
        command (str): The command to be executed in the shell.

    Returns:
        int: The return code of the command executed.
    """
    process = subprocess.Popen(command, stdout=subprocess.PIPE, **kwargs)
    while True:
        if process.poll() is not None:
            break
        output = process.stdout.readline()
        if output:
            print(output.strip())
    rc = process.poll()
    return rc


def fileToList(filename: str, strconv: callable = lambda x: x) -> list:
    """
    loads an input file with a\\n b\\n.. into a list [a,b,..]

    Args:
        filename (str): The path to the file to be loaded.
        strconv (callable): Function to convert each line. Defaults to identity function.

    Returns:
        list: The list of converted values from the file.
    """
    with open(filename) as f:
        return [strconv(val[:-1]) for val in f.readlines()]


def listToFile(
    li: List[str], filename: str, strconv: callable = lambda x: str(x)
) -> None:
    """
    listToFile loads a list with [a,b,..] into an input file a\\n b\\n..

    Args:
        li (list): The list of elements to be written to the file.
        filename (str): The name of the file where the list will be written.
        strconv (callable, optional): A function to convert each element of the list to a string. Defaults to str.

    Returns:
        None
    """
    with open(filename, "w") as f:
        for item in li:
            f.write("%s\n" % strconv(item))


def set_seed(seed: int = 42):
    """set random seed."""
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    # if n_gpu > 0:
    #     torch.cuda.manual_seed_all(seed)


def createFoldersFor(filepath):
    """
    will recursively create folders if needed until having all the folders required to save the file in this filepath
    """
    prevval = ""
    for val in os.path.expanduser(filepath).split("/")[:-1]:
        prevval += val + "/"
        if not os.path.exists(prevval):
            os.mkdir(prevval)


def category_str2int(category_strs: List[str]) -> List[int]:
    """
    category_str2int converts a list of category strings to a list of category integers.

    Args:
        category_strs (List[str]): A list of category strings to be converted.

    Returns:
        List[int]: A list of integers corresponding to the input category strings.
    """
    set_category_strs = set(category_strs)
    name2id = {name: i for i, name in enumerate(set_category_strs)}
    return [name2id[name] for name in category_strs]


def isnotebook() -> bool:
    """check whether excuting in jupyter notebook."""
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return True  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def get_free_gpu() -> int:
    """
    get_free_gpu finds the GPU with the most free memory using nvidia-smi.

    Returns:
        int: The index of the GPU with the most free memory.
    """
    import subprocess
    import sys
    from io import StringIO

    gpu_stats = subprocess.check_output(
        [
            "nvidia-smi",
            "--format=csv",
            "--query-gpu=memory.used,memory.free",
        ]
    ).decode("utf-8")
    gpu_df = pd.read_csv(
        StringIO(gpu_stats), names=["memory.used", "memory.free"], skiprows=1
    )
    print("GPU usage:\n{}".format(gpu_df))
    gpu_df["memory.free"] = gpu_df["memory.free"].map(lambda x: int(x.rstrip(" [MiB]")))
    idx = gpu_df["memory.free"].idxmax()
    print(
        "Find free GPU{} with {} free MiB".format(idx, gpu_df.iloc[idx]["memory.free"])
    )

    return idx


def get_git_commit() -> str:
    """
    get_git_commit gets the current git commit hash.

    Returns:
        str: The current git commit
    """
    return subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("utf-8").strip()


def ensure_dir(dirname):
    dirname = Path(dirname)
    if not dirname.is_dir():
        dirname.mkdir(parents=True, exist_ok=False)


def read_json(fname):
    fname = Path(fname)
    with fname.open("rt") as handle:
        return json.load(handle, object_hook=OrderedDict)


def write_json(content, fname):
    fname = Path(fname)
    with fname.open("wt") as handle:
        json.dump(content, handle, indent=4, sort_keys=False)


def inf_loop(data_loader):
    """wrapper function for endless data loader."""
    for loader in repeat(data_loader):
        yield from loader


def prepare_device(n_gpu_use):
    """
    setup GPU device if available. get gpu device indices which are used for DataParallel
    """
    n_gpu = torch.cuda.device_count()
    if n_gpu_use > 0 and n_gpu == 0:
        print(
            "Warning: There's no GPU available on this machine,"
            "training will be performed on CPU."
        )
        n_gpu_use = 0
    if n_gpu_use > n_gpu:
        print(
            f"Warning: The number of GPU's configured to use is {n_gpu_use}, but only {n_gpu} are "
            "available on this machine."
        )
        n_gpu_use = n_gpu
    device = torch.device("cuda:0" if n_gpu_use > 0 else "cpu")
    list_ids = list(range(n_gpu_use))
    return device, list_ids


"""Helper functions related to subsetting AnnData objects based on the file format 
specifications in the .config.vsh.yaml and slot mapping overrides."""


# create new anndata objects according to api spec
def subset_h5ad_by_format(adata, config, arg_name, field_rename_dict={}):
    """Create new anndata object according to slot info specifications.

    Arguments:
    adata -- An AnnData object to subset (required)
    config -- A Viash config object as read by openproblems.project.read_viash_config (required)
    arg_name -- The name of the argument in the config file that specifies the output format (required)
    field_rename_dict -- A mapping between the slots of the source h5ad and the slots of the destination h5ad.
      Example of slot_mapping:
        ```
        slot_mapping = {
          "layers": {
            "counts": par["layer_counts"],
          },
          "obs": {
            "cell_type": par["obs_cell_type"],
            "batch": par["obs_batch"],
          }
        }
    """
    import anndata as ad
    import pandas as pd

    assert isinstance(adata, ad.AnnData), "adata must be an AnnData object"
    assert isinstance(config, dict), "config must be a dictionary"

    # find argument
    arg = next(
        (x for x in config["all_arguments"] if x["clean_name"] == arg_name), None
    )
    assert arg, f"Argument '{arg_name}' not found in config"

    # find file format
    file_format = (arg.get("info") or {}).get("format")
    assert file_format, f"Argument '{arg_name}' has no .info.format"

    # find file format type
    file_format_type = file_format.get("type")
    assert file_format_type == "h5ad", "format must be a h5ad type"

    structs = ["layers", "obs", "var", "uns", "obsp", "obsm", "varp", "varm"]
    kwargs = {}

    for struct in structs:
        struct_format = file_format.get(struct, {})
        struct_rename = field_rename_dict.get(struct, {})

        # fetch data from adata
        data = {}
        for field_format in struct_format:
            dest_name = field_format["name"]
            # where to find the data. if the dest_name is in the rename dict, use the renamed name
            # as the source name, otherwise use the dest_name as the source name
            src_name = struct_rename.get(dest_name, dest_name)
            data[dest_name] = getattr(adata, struct)[src_name]

        if len(data) > 0:
            if struct in ["obs", "var"]:
                data = pd.concat(data, axis=1)
            kwargs[struct] = data
        elif struct in ["obs", "var"]:
            # if no columns need to be copied, we still need an 'obs' and a 'var'
            # to help determine the shape of the adata
            kwargs[struct] = getattr(adata, struct).iloc[:, []]

    return ad.AnnData(**kwargs)


def volcano(
    data,
    folder="",
    tohighlight=None,
    tooltips=[("gene", "@gene_id")],
    title="volcano plot",
    xlabel="log-fold change",
    ylabel="-log(Q)",
    maxvalue=100,
    searchbox=False,
    logfoldtohighlight=0.15,
    pvaltohighlight=0.1,
    showlabels=False,
):
    """
    Make an interactive volcano plot from Differential Expression analysis tools outputs

    Args:
    -----
        data: a df with rows genes and cols [log2FoldChange, pvalue, gene_id]
        folder: str of location where to save the plot, won't save if empty
        tohighlight: list[str] of genes to highlight in the plot
        tooltips: list[tuples(str,str)] if user wants tot specify another bokeh tooltip
        title: str plot title
        xlabel: str if user wants to specify the title of the x axis
        ylabel: str if user wants tot specify the title of the y axis
        maxvalue: float the max -log2(pvalue authorized usefull when managing inf vals)
        searchbox: bool whether or not to add a searchBox to interactively highlight genes
        logfoldtohighlight: float min logfoldchange when to diplay points
        pvaltohighlight: float min pvalue when to diplay points
        showlabels: bool whether or not to show a text above each datapoint with its label information

    Returns:
    --------
        The bokeh object
    """
    to_plot_not, to_plot_yes = selector(
        data,
        tohighlight if tohighlight is not None else [],
        logfoldtohighlight,
        pvaltohighlight,
    )
    hover = HoverTool(tooltips=tooltips, name="circles")

    # Create figure
    p = figure(title=title, width=650, height=450)

    p.xgrid.grid_line_color = "white"
    p.ygrid.grid_line_color = "white"
    p.xaxis.axis_label = xlabel
    p.yaxis.axis_label = ylabel

    # Add the hover tool
    p.add_tools(hover)
    p, source1 = add_points(
        p, to_plot_not, "log2FoldChange", "pvalue", color="#1a9641", maxvalue=maxvalue
    )
    p, source2 = add_points(
        p,
        to_plot_yes,
        "log2FoldChange",
        "pvalue",
        color="#fc8d59",
        alpha=0.6,
        outline=True,
        maxvalue=maxvalue,
    )
    if showlabels:
        labels = LabelSet(
            x="log2FoldChange",
            y="transformed_q",
            text_font_size="7pt",
            text="gene_id",
            level="glyph",
            x_offset=5,
            y_offset=5,
            source=source2,
            # renderers="canvas",
        )
        p.add_layout(labels)
    if searchbox:
        text = TextInput(title="text", value="gene")
        text.js_on_change(
            "value",
            CustomJS(
                args=dict(source=source1),
                code="""
                var data = source.data
                var value = cb_obj.value
                var gene_id = data.gene_id
                var a = -1
                for (let i=0; i < gene_id.length; i++) {
                    if ( gene_id[i]===value ) { a=i; console.log(i); data.size[i]=7; data.alpha[i]=1; data.color[i]='#fc8d59' }
                }
                source.data = data
                console.log(source)
                console.log(cb_obj)
                source.change.emit()
                console.log(source)
                """,
            ),
        )
        p = column(text, p)
    p.output_backend = "svg"
    if folder:
        save(p, folder + title.replace(" ", "_") + "_volcano.html")
        try:
            p.output_backend = "svg"
            export_svg(p, filename=folder + title.replace(" ", "_") + "_volcano.svg")
        except (RuntimeError, Exception) as e:
            print(f"Could not save SVG: {e}")
    try:
        show(p)
    except Exception as e:
        print(f"Could not show plot: {e}")
    return p


def add_points(p, df1, x, y, color="blue", alpha=0.2, outline=False, maxvalue=100):
    """parts of volcano plot"""
    # Define colors in a dictionary to access them with
    # the key from the pandas groupby funciton.
    df = df1.copy()
    transformed_q = -df[y].apply(np.log10).values
    transformed_q[transformed_q == np.inf] = maxvalue
    transformed_q[transformed_q > maxvalue] = maxvalue
    df["transformed_q"] = transformed_q
    df["color"] = color
    df["alpha"] = alpha
    df["size"] = 7
    source1 = ColumnDataSource(df)

    # Specify data source
    p.scatter(
        x=x,
        y="transformed_q",
        size="size",
        alpha="alpha",
        source=source1,
        color="color",
        name="circles",
    )
    if outline:
        p.scatter(
            x=x,
            y="transformed_q",
            size=7,
            alpha=1,
            source=source1,
            color="black",
            fill_color=None,
            name="outlines",
        )

    # prettify
    p.background_fill_color = "#DFDFE5"
    p.background_fill_alpha = 0.5
    return p, source1


def selector(
    df,
    valtoextract=[],
    logfoldtohighlight=0.15,
    pvaltohighlight=0.1,
    minlogfold=0.15,
    minpval=0.1,
):
    """Part of Volcano plot: A function to separate tfs from everything else"""
    toshow = (df.pvalue < minpval) & (abs(df.log2FoldChange) > minlogfold)
    df = df[toshow]
    sig = (df.pvalue < pvaltohighlight) & (abs(df.log2FoldChange) > logfoldtohighlight)
    if valtoextract:
        not_tf = ~df.gene_id.isin(valtoextract)
        is_tf = df.gene_id.isin(valtoextract)
        to_plot_not = df[~sig | not_tf]
        to_plot_yes = df[sig & is_tf]
    else:
        to_plot_not = df[~sig]
        to_plot_yes = df[sig]
    return to_plot_not, to_plot_yes


def correlationMatrix(
    data,
    names,
    colors=None,
    pvals=None,
    maxokpval=10**-9,
    other=None,
    title="correlation Matrix",
    dataIsCorr=False,
    invert=False,
    size=40,
    folder="",
    interactive=False,
    maxval=None,
    minval=None,
):
    """
    Make an interactive correlation matrix from an array using bokeh

    Args:
    -----
      data: arrayLike of int / float/ bool of size(names*val) or (names*names)
      names: list[str] of names for each rows
      colors: list[int] of size(names) a color for each names (good to display clusters)
      pvals: arraylike of int / float/ bool of size(names*val) or (names*names) with the corresponding pvalues
      maxokpval: float threshold when pvalue is considered good. otherwise lowers the size of the square
        until 10**-3 when it disappears
      other: arrayLike of int / float/ bool of size(names*val) or (names*names), an additional information
        matrix that you want ot display with opacity whereas correlations willl be displayed with
      title: str the plot title
      dataIsCorr: bool if not true, we will compute the corrcoef of the data array
      invert: bool whether or not to invert the matrix before running corrcoef
      size: int the plot size
      folder: str of folder location where to save the plot, won't save if empty
      interactive: bool whether or not to make the plot interactive (else will use matplotlib)
      maxval: float clamping coloring up to maxval
      minval: float clamping coloring down to minval

    Returns:
    -------
      the bokeh object if interactive else None

    """
    if not dataIsCorr:
        print("computing correlations")
        data = np.corrcoef(np.array(data) if not invert else np.array(data).T)
    else:
        data = np.array(data)
    regdata = data.copy()
    if minval is not None:
        data[data < minval] = minval
    if maxval is not None:
        data[data > maxval] = maxval
    data = data / data.max()
    TOOLS = (
        "hover,crosshair,pan,wheel_zoom,zoom_in,zoom_out,box_zoom,undo,redo,reset,save"
    )
    xname = []
    yname = []
    color = []
    alpha = []
    height = []
    width = []
    if type(colors) is list:
        print("we are assuming you want to display clusters with colors")
    elif other is not None:
        print(
            "we are assuming you want to display the other of your correlation with opacity"
        )
    if pvals is not None:
        print(
            "we are assuming you want to display the pvals of your correlation with size"
        )
        regpvals = pvals.copy()
        u = pvals < maxokpval
        pvals[~u] = np.log10(1 / pvals[~u])
        pvals = pvals / pvals.max()
        pvals[u] = 1
    if interactive:
        xname = []
        yname = []
        color = []
        for i, name1 in enumerate(names):
            for j, name2 in enumerate(names):
                xname.append(name1)
                yname.append(name2)
                if pvals is not None:
                    height.append(max(0.1, min(0.9, pvals[i, j])))
                    color.append(cc.coolwarm[int((data[i, j] * 127) + 127)])
                    alpha.append(min(abs(data[i, j]), 0.9))
                elif other is not None:
                    color.append(cc.coolwarm[int((data[i, j] * 127) + 127)])
                    alpha.append(
                        max(min(other[i, j], 0.9), 0.1) if other[i, j] != 0 else 0
                    )
                else:
                    alpha.append(min(abs(data[i, j]), 0.9))
                if colors is not None:
                    if type(colors) is list:
                        if colors[i] == colors[j]:
                            color.append(Category10[10][colors[i]])
                        else:
                            color.append("lightgrey")

                elif pvals is None and other is None:
                    color.append("grey" if data[i, j] > 0 else Category20[3][2])
        print(regdata.max())
        if pvals is not None:
            width = height.copy()
            data = dict(
                xname=xname,
                yname=yname,
                colors=color,
                alphas=alpha,
                data=regdata.ravel(),
                pvals=regpvals.ravel(),
                width=width,
                height=height,
            )
        else:
            data = dict(
                xname=xname, yname=yname, colors=color, alphas=alpha, data=data.ravel()
            )
        tt = [("names", "@yname, @xname"), ("value", "@data")]
        if pvals is not None:
            tt.append(("pvals", "@pvals"))
        p = figure(
            title=title if title is not None else "Correlation Matrix",
            x_axis_location="above",
            tools=TOOLS,
            x_range=list(reversed(names)),
            y_range=names,
            tooltips=tt,
        )

        p.width = 800
        p.height = 800
        p.grid.grid_line_color = None
        p.axis.axis_line_color = None
        p.axis.major_tick_line_color = None
        p.axis.major_label_text_font_size = "5pt"
        p.axis.major_label_standoff = 0
        p.xaxis.major_label_orientation = np.pi / 3
        p.output_backend = "svg"
        p.rect(
            "xname",
            "yname",
            width=0.9 if not width else "width",
            height=0.9 if not height else "height",
            source=data,
            color="colors",
            alpha="alphas",
            line_color=None,
            hover_line_color="black",
            hover_color="colors",
        )
        save(p, folder + title.replace(" ", "_") + "_correlation.html")
        try:
            p.output_backend = "svg"
            export_svg(
                p, filename=folder + title.replace(" ", "_") + "_correlation.svg"
            )
        except (RuntimeError, Exception) as e:
            print(f"Could not save SVG: {e}")
        try:
            show(p)
        except Exception as e:
            print(f"Could not show plot: {e}")
        return p  # show the plot
    else:
        plt.figure(figsize=(size, 200))
        plt.title("the correlation matrix")
        plt.imshow(data)
        plt.savefig(title + "_correlation.pdf")
        plt.show()


def heatmap(
    data,
    colors=None,
    title="correlation Matrix",
    size=40,
    other=None,
    folder="",
    interactive=False,
    pvals=None,
    maxokpval=10**-9,
    maxval=None,
    minval=None,
):
    """
    Make an interactive heatmap from a dataframe using bokeh

    Args:
    -----
      data: dataframe of int / float/ bool of size(names1*names2)
      colors: list[int] of size(names) a color for each names (good to display clusters)
      pvals: arraylike of int / float/ bool of size(names*val) or (names*names) with the corresponding pvalues
      maxokpval: float threshold when pvalue is considered good. otherwise lowers the size of the square
        until 10**-3 when it disappears
      title: str the plot title
      size: int the plot size
      folder: str of folder location where to save the plot, won't save if empty
      interactive: bool whether or not to make the plot interactive (else will use matplotlib)
      maxval: float clamping coloring up to maxval
      minval: float clamping coloring down to minval

    Returns:
    -------
      the bokeh object if interactive else None

    """
    regdata = data.copy()
    if minval is not None:
        data[data < minval] = minval
    if maxval is not None:
        data[data > maxval] = maxval
    data = data / data.max()
    data = data.values
    TOOLS = (
        "hover,crosshair,pan,wheel_zoom,zoom_in,zoom_out,box_zoom,undo,redo,reset,save"
    )
    xname = []
    yname = []
    color = []
    alpha = []
    height = []
    width = []
    if pvals is not None:
        print(
            "we are assuming you want to display the pvals of your correlation with size"
        )
        regpvals = pvals.copy()
        u = pvals < maxokpval
        pvals[~u] = np.log10(1 / pvals[~u])
        pvals = pvals / pvals.max()
        pvals[u] = 1
    if interactive:
        xname = []
        yname = []
        color = []
        for i, name1 in enumerate(regdata.index):
            for j, name2 in enumerate(regdata.columns):
                xname.append(name2)
                yname.append(name1)
                if pvals is not None:
                    # import pdb;pdb.set_trace()
                    height.append(max(0.1, min(0.9, pvals.loc[name1][name2])))
                    color.append(cc.coolwarm[int((data[i, j] * 128) + 127)])
                    alpha.append(min(abs(data[i, j]), 0.9))
                elif other is not None:
                    color.append(cc.coolwarm[int((data[i, j] * 128) + 127)])
                    alpha.append(
                        max(min(other[i, j], 0.9), 0.1) if other[i, j] != 0 else 0
                    )
                else:
                    alpha.append(min(abs(data[i, j]), 0.9))
                if colors is not None:
                    if type(colors) is list:
                        if colors[i] == colors[j]:
                            color.append(Category10[10][colors[i]])
                        else:
                            color.append("lightgrey")

                elif pvals is None and other is None:
                    color.append("grey" if data[i, j] > 0 else Category20[3][2])
        if pvals is not None:
            width = height.copy()
            data = dict(
                xname=xname,
                yname=yname,
                colors=color,
                alphas=alpha,
                data=regdata.values.ravel(),
                pvals=regpvals.values.ravel(),
                width=width,
                height=height,
            )
        else:
            data = dict(
                xname=xname, yname=yname, colors=color, alphas=alpha, data=data.ravel()
            )
        tt = [("names", "@yname, @xname"), ("value", "@data")]
        if pvals is not None:
            tt.append(("pvals", "@pvals"))
        p = figure(
            title=title if title is not None else "Heatmap",
            x_axis_location="above",
            tools=TOOLS,
            x_range=list(reversed(regdata.columns.astype(str).tolist())),
            y_range=regdata.index.tolist(),
            tooltips=tt,
        )

        p.width = 800
        p.height = 800
        p.grid.grid_line_color = None
        p.axis.axis_line_color = None
        p.axis.major_tick_line_color = None
        p.axis.major_label_text_font_size = "5pt"
        p.axis.major_label_standoff = 0
        p.xaxis.major_label_orientation = np.pi / 3
        p.output_backend = "svg"
        p.rect(
            "xname",
            "yname",
            width=0.9 if not width else "width",
            height=0.9 if not height else "height",
            source=data,
            color="colors",
            alpha="alphas",
            line_color=None,
            hover_line_color="black",
            hover_color="colors",
        )
        save(p, folder + title.replace(" ", "_") + "_heatmap.html")
        try:
            p.output_backend = "svg"
            export_svg(
                p, filename=folder + title.replace(" ", "_") + "_correlation.svg"
            )
        except (RuntimeError, Exception) as e:
            print(f"Could not save SVG: {e}")
        try:
            show(p)
        except Exception as e:
            print(f"Could not show plot: {e}")
        return p  # show the plot
    else:
        plt.figure(figsize=size)
        plt.title("the correlation matrix")
        plt.imshow(data)
        plt.savefig(title + "_correlation.pdf")
        plt.show()



import torch

# Adapted from https://github.com/gpeyre/SinkhornAutoDiff
# and from https://github.com/dfdazac/wassdistance/blob/master/layers.py


class SinkhornDistance(torch.nn.Module):
    def __init__(self, eps: float = 1e-2, max_iter: int = 100, reduction: str = "none"):
        """
        SinkhornDistance Initialize the SinkhornDistance class

        Args:
            eps (float, optional): Regularization parameter. Defaults to 1e-2.
            max_iter (int, optional): Maximum number of Sinkhorn iterations. Defaults to 100.
            reduction (str, optional): Specifies the reduction to apply to the output. Defaults to "none".
        """
        super(SinkhornDistance, self).__init__()
        self.eps = eps
        self.max_iter = max_iter
        self.reduction = reduction

    def forward(self, c: torch.Tensor) -> torch.Tensor:
        """
        forward Compute the Sinkhorn distance between two measures with cost matrix c

        Args:
            c (torch.Tensor): The cost matrix between the two measures.

        Returns:
            torch.Tensor: The computed Sinkhorn distance.
        """
        C = -c
        x_points = C.shape[-2]
        batch_size = C.shape[0]

        # both marginals are fixed with equal weights
        mu = (
            torch.empty(
                batch_size,
                x_points,
                dtype=C.dtype,
                requires_grad=False,
                device=C.device,
            )
            .fill_(1.0 / x_points)
            .squeeze()
        )
        nu = (
            torch.empty(
                batch_size,
                x_points,
                dtype=C.dtype,
                requires_grad=False,
                device=C.device,
            )
            .fill_(1.0 / x_points)
            .squeeze()
        )
        u = torch.zeros_like(mu)
        v = torch.zeros_like(nu)

        # Stopping criterion
        thresh = 1e-12

        # Sinkhorn iterations
        for i in range(self.max_iter):
            if i % 2 == 0:
                u1 = u  # useful to check the update
                u = (
                    self.eps
                    * (torch.log(mu) - torch.logsumexp(self.M(C, u, v), dim=-1))
                    + u
                )
                err = (u - u1).abs().sum(-1).mean()
            else:
                v = (
                    self.eps
                    * (
                        torch.log(nu)
                        - torch.logsumexp(self.M(C, u, v).transpose(-2, -1), dim=-1)
                    )
                    + v
                )
                v = v.detach().requires_grad_(False)
                v[v > 9 * 1e8] = 0.0
                v = v.detach().requires_grad_(True)

            if err.item() < thresh:
                break

        U, V = u, v
        # Transport plan pi = diag(a)*K*diag(b)
        pi = torch.exp(self.M(C, U, V))

        # Sinkhorn distance

        return pi, C, U, V

    def M(self, C, u, v):
        "Modified cost for logarithmic updates"
        """$M_{ij} = (-c_{ij} + u_i + v_j) / epsilon$"""
        return (-C + u.unsqueeze(-1) + v.unsqueeze(1)) / self.eps

    @staticmethod
    def ave(u, u1, tau):
        "Barycenter subroutine, used by kinetic acceleration through extrapolation."
        return tau * u + (1 - tau) * u1



import os
from typing import Any, Dict, List, Optional

import bionty as bt
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
import torch
import torch.nn.functional as F
from anndata import AnnData, concat
from scdataloader import Collator, Preprocessor
from scdataloader.data import SimpleAnnDataset
from scdataloader.utils import get_descendants, random_str
from scib_metrics.benchmark import Benchmarker
from scipy.stats import spearmanr
from simpler_flash import FlashTransformer
from sklearn.metrics import f1_score
from torch.utils.data import DataLoader
from tqdm import tqdm

from scprint2.model import loss

FILE_LOC = os.path.dirname(os.path.realpath(__file__))


class FinetuneBatchClass:
    def __init__(
        self,
        batch_key: str = "batch",
        predict_keys: List[str] = ["cell_type_ontology_term_id"],
        max_len: int = 5000,
        learn_batches_on: Optional[str] = None,
        num_workers: int = 8,
        batch_size: int = 16,
        num_epochs: int = 8,
        do_mmd_on: Optional[str] = None,
        lr: float = 0.0002,
        ft_mode: str = "xpressor",
        frac_train: float = 0.8,
        loss_scalers: dict = {},
        use_knn: bool = True,
    ):
        """
        Embedder a class to embed and annotate cells using a model

        Args:
            batch_key (str, optional): The key in adata.obs that indicates the batch information. Defaults to "batch".
            learn_batches_on (str, optional): The key in adata.obs to learn batch embeddings on. Defaults to None.
                if none, will not learn the batch embeddings.
                the goal is e.g. when having a new species, to learn an embedding for it during finetuning and replace
                the "learn_batches_on" embedding in the model with it, in this case it should be "organism_ontology_term_id".
                batch correction might indeed be better learnt with this additional argument in some cases.
            do_mmd_on (str, optional):The key in adata.obs to learn batch embeddings on. Defaults to None.
                this embedding should have less batch information in it, after finetuning.
            predict_keys (List[str], optional): List of keys in adata.obs to predict during fine-tuning. Defaults to ["cell_type_ontology_term_id"].
            batch_size (int, optional): The size of the batches to be used in the DataLoader. Defaults to 64.
            num_workers (int, optional): The number of worker processes to use for data loading. Defaults to 8.
            max_len (int, optional): The maximum length of the sequences to be processed. Defaults to 5000.
            lr (float, optional): The learning rate for the optimizer. Defaults to 0.0002.
            num_epochs (int, optional): The number of epochs to train the model. Defaults to 8.
            ft_mode (str, optional): The fine-tuning mode, either "xpressor" or "full". Defaults to "xpressor".
            frac_train (float, optional): The fraction of data to be used for training. Defaults to 0.8.
            loss_scalers (dict, optional): A dictionary specifying the scaling factors for different loss components. Defaults to {}.
                expr, class, mmd, kl, and any of the predict_keys can be specified.
            use_knn (bool, optional): Whether to use k-nearest neighbors information. Defaults to True.
        """
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.batch_key = batch_key
        self.learn_batches_on = learn_batches_on
        self.predict_keys = predict_keys
        self.max_len = max_len
        self.lr = lr
        self.num_epochs = num_epochs
        self.ft_mode = ft_mode
        self.frac_train = frac_train
        self.batch_emb = None
        self.batch_encoder = {}
        self.do_mmd_on = do_mmd_on
        self.loss_scalers = loss_scalers
        self.use_knn = use_knn

    def __call__(
        self,
        model: torch.nn.Module,
        adata: AnnData = None,
        train_data: AnnData = None,
        val_data: AnnData = None,
    ) -> torch.nn.Module:
        """
        __call__ function to call the embedding

        Args:
            model (torch.nn.Module): The scPRINT model to be used for embedding and annotation.
            adata (AnnData): The annotated data matrix of shape n_obs x n_vars. Rows correspond to cells and columns to genes.
                Defaults to None.
                if provided, it will be split into training and validation sets.
            train_data (AnnData, optional): The training data. Defaults to None.
                if adata is provided, this will be ignored.
            val_data (AnnData, optional): The validation data. Defaults to None.
                if adata is provided, this will be ignored.

        Raises:
            ValueError: If the model does not have a logger attribute.
            ValueError: If the model does not have a global_step attribute.

        Returns:
            torch.nn.Module: the fine-tuned model
        """
        # one of "all" "sample" "none"
        model.predict_mode = "none"
        if self.ft_mode == "xpressor":
            for val in model.parameters():
                val.requires_grad = False
                # setting all to TRUE

            for val in model.cell_transformer.parameters():
                val.requires_grad = True
            for val in model.transformer.blocks[-1].parameters():
                val.requires_grad = True
            for i in model.transformer.blocks:
                i.cross_attn.requires_grad = True
            for val in model.compressor.parameters():
                val.requires_grad = True
            for val in self.predict_keys:
                for val in model.cls_decoders[val].parameters():
                    val.requires_grad = True
        elif self.ft_mode == "full":
            for val in model.parameters():
                val.requires_grad = True
        else:
            raise ValueError("ft_mode must be one of 'xpressor' or 'full'")

        # PREPARING THE DATA
        if adata is not None:
            n_train = int(self.frac_train * len(adata))
            train_idx = np.random.choice(len(adata), n_train, replace=False)
            val_idx = np.setdiff1d(np.arange(len(adata)), train_idx)

            train_data = adata[train_idx].copy()
            val_data = adata[val_idx].copy()

            print(f"Training data: {train_data.shape}")
            print(f"Validation data: {val_data.shape}")

        mencoders = {}
        for k, v in model.label_decoders.items():
            mencoders[k] = {va: ke for ke, va in v.items()}
        # this needs to remain its original name as it is expect like that by collator, otherwise need to send org_to_id as params

        for i in self.predict_keys:
            if len(set(train_data.obs[i]) - set(mencoders[i].keys())) > 0:
                print("missing labels for ", i)
                train_data.obs[i] = train_data.obs[i].apply(
                    lambda x: x if x in mencoders[i] else "unknown"
                )
        if "organism_ontology_term_id" not in self.predict_keys:
            self.predict_keys.append("organism_ontology_term_id")
        # create datasets
        self.batch_encoder = {
            i: n
            for n, i in enumerate(
                train_data.obs[self.batch_key].astype("category").cat.categories
            )
        }
        mencoders[self.batch_key] = self.batch_encoder
        train_dataset = SimpleAnnDataset(
            train_data,
            obs_to_output=self.predict_keys + [self.batch_key],
            get_knn_cells=model.expr_emb_style == "metacell" and self.use_knn,
            encoder=mencoders,
        )
        if val_data is not None:
            for i in self.predict_keys:
                if i != "organism_ontology_term_id":
                    if len(set(val_data.obs[i]) - set(mencoders[i].keys())) > 0:
                        val_data.obs[i] = val_data.obs[i].apply(
                            lambda x: x if x in mencoders[i] else "unknown"
                        )
            self.batch_encoder.update(
                {
                    i: n + len(self.batch_encoder)
                    for n, i in enumerate(
                        val_data.obs[self.batch_key].astype("category").cat.categories
                    )
                    if i not in self.batch_encoder
                }
            )
            mencoders[self.batch_key] = self.batch_encoder
            val_dataset = SimpleAnnDataset(
                val_data,
                obs_to_output=self.predict_keys + [self.batch_key],
                get_knn_cells=model.expr_emb_style == "metacell" and self.use_knn,
                encoder=mencoders,
            )

        # Create collator
        collator = Collator(
            organisms=model.organisms,
            valid_genes=model.genes,
            class_names=self.predict_keys + [self.batch_key],
            how="random expr",  # or "all expr" for full expression
            max_len=self.max_len,
            org_to_id=mencoders.get("organism_ontology_term_id", {}),
        )

        # Create data loaders
        train_loader = DataLoader(
            train_dataset,
            collate_fn=collator,
            batch_size=self.batch_size,  # Adjust based on GPU memory
            num_workers=self.num_workers,
            shuffle=True,
        )
        if val_data is not None:
            val_loader = DataLoader(
                val_dataset,
                collate_fn=collator,
                batch_size=self.batch_size,
                num_workers=self.num_workers,
                shuffle=False,
            )

        if self.learn_batches_on is not None:
            if val_data is not None:
                print(
                    "all batch key values in val_data should also be present in train_adata!!!"
                )
            self.batch_emb = torch.nn.Embedding(
                num_embeddings=train_data.obs[self.batch_key].nunique(),
                embedding_dim=(
                    model.compressor[self.learn_batches_on].fc_mu.weight.shape[0]
                    if hasattr(model, "compressor")
                    else model.d_model
                ),
            )

        ## PREPARING THE OPTIM
        all_params = (
            list(model.parameters())
            # + list(batch_cls.parameters())
            + (
                list(self.batch_emb.parameters())
                if self.learn_batches_on is not None
                else []
            )
        )

        # Setup optimizer
        optimizer = torch.optim.AdamW(
            all_params,
            lr=self.lr,
            weight_decay=0.01,
            betas=(0.9, 0.999),
            eps=1e-8,
        )

        # Setup scheduler
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer, mode="min", factor=0.5, patience=2
        )

        # Setup automatic mixed precision
        scaler = torch.cuda.amp.GradScaler() if torch.cuda.is_available() else None

        for k, i in model.mat_labels_hierarchy.items():
            model.mat_labels_hierarchy[k] = i.to(model.device)

        ## train
        for epoch in range(self.num_epochs):
            print(f"\nEpoch {epoch + 1}/{self.num_epochs}")
            print(f"Current learning rate: {optimizer.param_groups[0]['lr']:.2e}")

            # Training phase
            train_loss = 0.0
            train_steps = 0
            avg_expr = 0
            avg_cls = 0
            avg_mmd = 0

            pbar = tqdm(train_loader, desc="Training")
            model.train()
            for batch_idx, batch in enumerate(pbar):
                optimizer.zero_grad()
                total_loss, cls_loss, mmd, loss_expr = self.batch_corr_pass(
                    batch, model
                )
                # Backward pass
                scaler.scale(total_loss).backward()
                scaler.unscale_(optimizer)
                torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=1.0)
                scaler.step(optimizer)
                scaler.update()

                train_loss += total_loss.item()
                train_steps += 1
                avg_cls += cls_loss.item()
                avg_expr += loss_expr.item()
                avg_mmd += mmd
                # Update progress bar
                pbar.set_postfix(
                    {
                        "loss": f"{total_loss.item():.4f}",
                        "avg_loss": f"{train_loss / train_steps:.4f}",
                        "cls_loss": f"{cls_loss.item():.4f}",
                        "mmd_loss": f"{mmd:.4f}",
                        "expr_loss": f"{loss_expr.item():.4f}",
                    }
                )

            # Validation phase
            if val_data is not None:
                model.eval()
                val_loss = 0.0
                val_steps = 0
                val_loss_expr = 0.0
                val_mmd = 0.0
                val_cls = 0.0
                val_loss_to_prt = 0.0

                with torch.no_grad():
                    for batch in val_loader:  # tqdm(val_loader, desc="Validation"):
                        loss_val, cls_loss, mmd, loss_expr = self.batch_corr_pass(
                            batch, model
                        )
                        val_loss_to_prt += loss_val.item()
                        val_loss += loss_val.item()
                        val_steps += 1
                        val_loss_expr += loss_expr.item()
                        val_mmd += mmd
                        val_cls += cls_loss.item()
                try:
                    avg_val_loss = val_loss_to_prt / val_steps
                    avg_train_loss = train_loss / train_steps
                except ZeroDivisionError:
                    print(
                        "Error: Division by zero occurred while calculating average losses."
                    )
                    avg_train_loss = 0
                print(
                    "cls_loss: {:.4f}, mmd_loss: {:.4f}, expr_loss: {:.4f}".format(
                        val_cls / val_steps,
                        val_mmd / val_steps,
                        val_loss_expr / val_steps,
                    )
                )
                print(f"Train Loss: {avg_train_loss:.4f}, Val Loss: {avg_val_loss:.4f}")

                # Store LR before scheduler step for comparison
                lr_before = optimizer.param_groups[0]["lr"]

                # Update learning rate
                scheduler.step(avg_val_loss)

                # Check if LR was reduced
                lr_after = optimizer.param_groups[0]["lr"]
                if lr_after < lr_before:
                    print(
                        f"🔻 Learning rate reduced from {lr_before:.2e} to {lr_after:.2e} (factor: {lr_after / lr_before:.3f})"
                    )
                else:
                    print(f"✅ Learning rate unchanged: {lr_after:.2e}")

                # Early stopping check (simple implementation)
                if epoch > 3 and val_loss / val_steps > 1.3 * avg_train_loss:
                    print("Early stopping due to overfitting")
                    break

        print("Manual fine-tuning completed!")
        model.eval()
        return model

    def batch_corr_pass(self, batch, model):
        gene_pos = batch["genes"].to(model.device)
        expression = batch["x"].to(model.device)
        depth = batch["depth"].to(model.device)
        class_elem = batch["class"].long().to(model.device)
        total_loss = 0
        # Forward pass with automatic mixed precisio^n
        with torch.cuda.amp.autocast():
            # Forward pass
            output = model.forward(
                gene_pos,
                expression,
                req_depth=depth,
                depth_mult=expression.sum(1),
                do_class=True,
                metacell_token=torch.zeros_like(depth),
            )
            ## adaptor on ct_emb
            # ctpos = model.classes.index("cell_type_ontology_term_id") + 1
            # emb = output["output_cell_embs"][:, ctpos, :]
            #
            # output["output_cell_embs"][:, ctpos, :] = adaptor_layer(
            #    torch.cat([emb, class_elem[:, 1].unsqueeze(1).float()], dim=1)
            # )
            if self.learn_batches_on is not None:
                batch_pos = model.classes.index(self.learn_batches_on) + 1
                output["output_cell_embs"][:, batch_pos, :] = self.batch_emb(
                    class_elem[:, -1]
                )

            ## generate expr loss
            output_gen = model._generate(
                cell_embs=output["output_cell_embs"],
                gene_pos=gene_pos,
                depth_mult=expression.sum(1),
                req_depth=depth,
            )
            if "zero_logits" in output_gen:
                loss_expr = loss.zinb(
                    theta=output_gen["disp"],
                    pi=output_gen["zero_logits"],
                    mu=output_gen["mean"],
                    target=expression,
                )
                if model.zinb_and_mse:
                    loss_expr += (
                        loss.mse(
                            input=torch.log(output_gen["mean"] + 1)
                            * (1 - torch.sigmoid(output_gen["zero_logits"])),
                            target=torch.log(expression + 1),
                        )
                        / 10  # scale to make it more similar to the zinb
                    )
            else:
                loss_expr = loss.mse(
                    input=torch.log(output_gen["mean"] + 1),
                    target=torch.log(expression + 1),
                )
            # Add expression loss to total
            total_loss += loss_expr * self.loss_scalers.get("expr", 0.5)

            # ct
            cls_loss = 0
            for clas in self.predict_keys:
                cls_output = output.get("cls_output_" + clas)
                # ct_output = output["output_cell_embs"][:, ctpos, :]
                # cls_output = model.cls_decoders["cell_type_ontology_term_id"](ct_output)
                cls_loss += loss.hierarchical_classification(
                    pred=cls_output,
                    cl=class_elem[:, self.predict_keys.index(clas)],
                    labels_hierarchy=(
                        model.mat_labels_hierarchy.get(clas).to("cuda")
                        if clas in model.mat_labels_hierarchy
                        else None
                    ),
                ) * self.loss_scalers.get(clas, 1)

            # organ class
            # org_emb = output["compressed_cell_embs"][
            #    model.classes.index("organism_ontology_term_id") + 1
            # ]
            # cls_loss += F.cross_entropy(
            #    input=batch_cls(org_emb),
            #    target=class_elem[:, 1],
            # )
            total_loss += cls_loss * self.loss_scalers.get("class", 1)
            tot_mmd = 0
            if self.do_mmd_on is not None:
                pos = model.classes.index(self.do_mmd_on) + 1
                # Apply gradient reversal to the input embedding
                selected_emb = (
                    output["compressed_cell_embs"][pos]
                    if model.compressor is not None
                    else output["input_cell_embs"][:, pos, :]
                )
                for i in set(class_elem[:, -1].cpu().numpy()):
                    if (class_elem[:, -1] == i).sum() < 2:
                        # need at least 2 samples to compute mmd
                        class_elem[class_elem[:, -1] == i, 1] = (
                            -1
                        )  # assign to dummy class
                    # compare each batch to all other batches
                for i in set(class_elem[:, -1].cpu().numpy()):
                    if i == -1:
                        continue
                    X, Y = (
                        selected_emb[class_elem[:, -1] == i],
                        selected_emb[class_elem[:, -1] != i],
                    )
                    mmd = mmd_loss(X, Y)
                    if torch.isnan(mmd):
                        print("mmd nan")
                    tot_mmd += mmd.item() if not torch.isnan(mmd) else 0
                # Add adversarial loss to total loss
                total_loss += tot_mmd * self.loss_scalers.get("mmd", 3)
            if "vae_kl_loss" in output:
                total_loss += output["vae_kl_loss"] * self.loss_scalers.get("kl", 0.5)
        return total_loss, cls_loss, tot_mmd, loss_expr


def mmd_loss(X: torch.Tensor, Y: torch.Tensor) -> torch.Tensor:
    """
    Compute Maximum Mean Discrepancy (MMD) loss between two 2D embedding matrices.

    Args:
        X (torch.Tensor): Tensor of shape (n1, emb_dim) - first set of embeddings
        Y (torch.Tensor): Tensor of shape (n2, emb_dim) - second set of embeddings

    Returns:
        torch.Tensor: MMD loss value (negative to encourage dissimilarity)
    """

    def rbf_kernel(x, y, sigma):
        """Compute RBF kernel between two sets of vectors"""
        distance = torch.cdist(x, y, p=2) ** 2
        return torch.exp(-distance / (2 * sigma**2))

    def energy_kernel(x, y):
        """Compute Energy kernel between two sets of vectors"""
        distance = torch.cdist(x, y, p=2)
        return -distance

    # Use multiple kernel bandwidths for better performance
    sigmas = [0]  # [0.1, 1.0, 10.0]
    mmd_loss = 0.0

    for sigma in sigmas:
        # K(X, X) - kernel matrix within first group (n1 x n1)
        # k_xx = rbf_kernel(X, X, sigma)
        k_xx = energy_kernel(X, X)
        # K(Y, Y) - kernel matrix within second group (n2 x n2)
        # k_yy = rbf_kernel(Y, Y, sigma)
        k_yy = energy_kernel(Y, Y)
        # K(X, Y) - kernel matrix between groups (n1 x n2)
        # k_xy = rbf_kernel(X, Y, sigma)
        k_xy = energy_kernel(X, Y)

        # Unbiased MMD estimation
        n1 = X.shape[0]
        n2 = Y.shape[0]

        # Remove diagonal elements for unbiased estimation of K(X,X) and K(Y,Y)
        # For K(X,X): exclude diagonal
        if n1 > 1:
            mask_xx = 1 - torch.eye(n1, device=X.device)
            k_xx_term = (k_xx * mask_xx).sum() / (n1 * (n1 - 1))
        else:
            k_xx_term = 0.0

        # For K(Y,Y): exclude diagonal
        if n2 > 1:
            mask_yy = 1 - torch.eye(n2, device=Y.device)
            k_yy_term = (k_yy * mask_yy).sum() / (n2 * (n2 - 1))
        else:
            k_yy_term = 0.0

        # For K(X,Y): use all elements (no diagonal to exclude)
        k_xy_term = k_xy.mean()

        # MMD^2 = E[K(X,X)] + E[K(Y,Y)] - 2*E[K(X,Y)]
        mmd_squared = k_xx_term + k_yy_term - 2 * k_xy_term
        mmd_loss += mmd_squared

    # Return negative MMD to encourage dissimilarity (higher MMD = more different)
    return mmd_loss / len(sigmas)


class FinetuneGRN:
    pass


class FinetuneGeneEmb:
    pass


class FinetuneNewClass:
    pass


class FinetuneUpdateClass:
    pass



import os

import numpy as np
import pandas as pd
import torch

# from RNABERT import RNABERT
from torch.nn import AdaptiveAvgPool1d
from tqdm import tqdm

from .. import utils
from .protein_embedder import ESM2


def protein_embeddings_generator(
    genedf: pd.DataFrame = None,
    organism: str = "homo_sapiens",  # mus_musculus,
    cache: bool = True,
    fasta_path: str = "/tmp/data/fasta/",
    embedding_size: int = 512,
    embedder: str = "esm3",  # or glm2
    cuda: bool = True,
):
    """
    protein_embeddings_generator embed a set of genes using fasta file and LLMs

    Args:
        genedf (pd.DataFrame): A DataFrame containing gene information.
        organism (str, optional): The organism to which the genes belong. Defaults to "homo_sapiens".
        cache (bool, optional): If True, the function will use cached data if available. Defaults to True.
        fasta_path (str, optional): The path to the directory where the fasta files are stored. Defaults to "/tmp/data/fasta/".
        embedding_size (int, optional): The size of the embeddings to be generated. Defaults to 512.
    Returns:
        pd.DataFrame: Returns a DataFrame containing the protein embeddings.
        pd.DataFrame: Returns the naming dataframe.
    """
    # given a gene file and organism
    # load the organism fasta if not already done
    fasta_path_pep, fasta_path_ncrna = utils.load_fasta_species(
        species=organism, output_path=fasta_path, cache=cache
    )
    # subset the fasta
    fasta_name = fasta_path_pep.split("/")[-1]
    utils.utils.run_command(["gunzip", fasta_path_pep])
    protgenedf = (
        genedf[genedf["biotype"] == "protein_coding"] if genedf is not None else None
    )
    found, naming_df = utils.subset_fasta(
        protgenedf.index.tolist() if protgenedf is not None else None,
        subfasta_path=fasta_path + "subset.fa",
        fasta_path=fasta_path + fasta_name[:-3],
        drop_unknown_seq=True,
    )
    if embedder == "esm2":
        prot_embedder = ESM2()
        prot_embeddings = prot_embedder(
            fasta_path + "subset.fa", output_folder=fasta_path + "esm_out/", cache=cache
        )
    elif embedder == "esm3":
        from Bio import SeqIO
        from esm.models.esmc import ESMC
        from esm.sdk.api import ESMProtein, LogitsConfig

        prot_embeddings = []
        names = []
        client = ESMC.from_pretrained("esmc_600m").to("cuda" if cuda else "cpu")
        conf = LogitsConfig(sequence=True, return_embeddings=True)
        with (open(fasta_path + "subset.fa", "r") as fasta,):
            for record in tqdm(SeqIO.parse(fasta, "fasta")):
                protein = ESMProtein(sequence=str(record.seq))
                protein_tensor = client.encode(protein)
                logits_output = client.logits(protein_tensor, conf)
                prot_embeddings.append(
                    logits_output.embeddings[0].mean(0).cpu().numpy().tolist()
                )
                names.append(record.id)
    else:
        raise ValueError(f"Embedder {embedder} not supported")
    # load the data and erase / zip the rest
    # utils.utils.run_command(["gzip", fasta_path + fasta_name[:-3]])
    # return the embedding and gene file
    # TODO: to redebug
    # do the same for RNA
    # rnagenedf = genedf[genedf["biotype"] != "protein_coding"]
    # fasta_file = next(
    #    file for file in os.listdir(fasta_path) if file.endswith(".ncrna.fa.gz")
    # )
    # utils.utils.run_command(["gunzip", fasta_path + fasta_file])
    # utils.subset_fasta(
    #    rnagenedf["ensembl_gene_id"].tolist(),
    #    subfasta_path=fasta_path + "subset.ncrna.fa",
    #    fasta_path=fasta_path + fasta_file[:-3],
    #    drop_unknown_seq=True,
    # )
    # rna_embedder = RNABERT()
    # rna_embeddings = rna_embedder(fasta_path + "subset.ncrna.fa")
    ## Check if the sizes of the cembeddings are not the same
    # utils.utils.run_command(["gzip", fasta_path + fasta_file[:-3]])
    #
    m = AdaptiveAvgPool1d(embedding_size)
    prot_embeddings = pd.DataFrame(
        data=m(torch.tensor(np.array(prot_embeddings))), index=names
    )
    # rna_embeddings = pd.DataFrame(
    #    data=m(torch.tensor(rna_embeddings.values)), index=rna_embeddings.index
    # )
    # Concatenate the embeddings
    return prot_embeddings, naming_df  # pd.concat([prot_embeddings, rna_embeddings])



EMBEDDING_DATASETS = {
    "lung": "https://figshare.com/ndownloader/files/24539942",
    "pancreas": "https://figshare.com/ndownloader/files/24539828",
    "kidney": "https://datasets.cellxgene.cziscience.com/01bc7039-961f-4c24-b407-d535a2a7ba2c.h5ad",
    # "gtex": "https://datasets.cellxgene.cziscience.com/661d5ec2-ca57-413c-8374-f49b0054ddba.h5ad",
    "bone_marrow_5batch": "https://datasets.cellxgene.cziscience.com/b2eca8f3-b461-45fd-8639-890bbbf050aa.h5ad",
}

DENOISE_DATASETS = {
    "intestine": "https://datasets.cellxgene.cziscience.com/d9a99b4a-3755-47c4-8eb5-09821ffbde17.h5ad",  # R4ZHoQegxXdSFNFY5LGe in my case # R4ZHoQegxXdSFNFYMaIQ
    "retina": "https://datasets.cellxgene.cziscience.com/53bd4177-79c6-40c8-b84d-ff300dcf1b5b.h5ad",  # gNNpgpo6gATjuxTE7CCp in my case # gNNpgpo6gATjuxTEW2mj
    "kidney": "https://datasets.cellxgene.cziscience.com/01bc7039-961f-4c24-b407-d535a2a7ba2c.h5ad",  # in my case d0JqVUfPuonxM3K3USwN
    "glio_smart_highdepth": "https://datasets.cellxgene.cziscience.com/6ec440b4-542a-4022-ac01-56f812e25593.h5ad",  # in my case s8x0Idi587LQtXCo0Pif
    "lung_smart": "https://datasets.cellxgene.cziscience.com/6ebba0e0-a159-406f-8095-451115673a2c.h5ad",  # in my case NwFOBmT6emiJaeyrEc66
}


def download_datasets():
    """
    Download the required test datasets in case you are in a server without internet access.
    """
    import scanpy as sc

    for dataset_name, url in EMBEDDING_DATASETS.items():
        sc.read(
            "../data/"
            + dataset_name
            + (".h5ad" if not dataset_name.endswith(".h5ad") else ""),
            backup_url=url,
        )
    for dataset_name, url in DENOISE_DATASETS.items():
        sc.read(
            "../data/"
            + dataset_name
            + (".h5ad" if not dataset_name.endswith(".h5ad") else ""),
            backup_url=url,
        )



import torch
from jsonargparse import class_from_function
from lightning.pytorch.callbacks import (
    EarlyStopping,
    LearningRateMonitor,
    ModelCheckpoint,
    StochasticWeightAveraging,
)
from lightning.pytorch.cli import LightningCLI, _get_short_description
from lightning.pytorch.trainer import Trainer

from scprint2.tasks import (
    Denoiser,
    Embedder,
    FinetuneBatchClass,
    GeneEmbeddingExtractor,
    GNInfer,
    Imputer,
)

from .trainer import TrainingMode

TASKS = [
    ("embed", Embedder),
    ("gninfer", GNInfer),
    ("denoise", Denoiser),
    ("finetune", FinetuneBatchClass),
    ("gene_emb", GeneEmbeddingExtractor),
    ("impute", Imputer),
]
TIMEOUT = 3600 * 9  # 9 hours in seconds


class MyCLI(LightningCLI):
    """
    MyCLI is a subclass of LightningCLI to add some missing params and create bindings between params of the model and the data.

    Used to allow calling denoise / embed / gninfer from the command line.
    Also to add more parameters and link parameters between the scdataloader and the scPRINT-2 model.
    """

    def add_arguments_to_parser(self, parser):
        parser.add_lightning_class_args(EarlyStopping, "scprint_early_stopping")
        parser.set_defaults(
            {
                "scprint_early_stopping.monitor": "val_loss",
                # patience is the number of consecutive epochs with no improvement after which learning rate will be reduced.
                "scprint_early_stopping.patience": 5,
                # Prevent early stopping from checking on train epoch end when val_loss isn't available yet
                "scprint_early_stopping.check_on_train_epoch_end": False,
            }
        )
        parser.add_lightning_class_args(
            LearningRateMonitor, "scprint_learning_rate_monitor"
        )
        parser.set_defaults({"scprint_learning_rate_monitor.logging_interval": "epoch"})
        parser.add_lightning_class_args(TrainingMode, "scprint_training")
        parser.link_arguments("data.genes_dict", "model.genes", apply_on="instantiate")
        parser.link_arguments(
            "data.decoders", "model.label_decoders", apply_on="instantiate"
        )
        parser.link_arguments(
            "data.labels_hierarchy", "model.labels_hierarchy", apply_on="instantiate"
        )
        parser.link_arguments(
            "data.n_bins", "model.n_input_bins", apply_on="instantiate"
        )
        parser.link_arguments("data.classes", "model.classes", apply_on="instantiate")
        parser.link_arguments(
            "data.organisms", "model.organisms", apply_on="instantiate"
        )

        parser.add_argument("--set_float32_matmul_precision", type=bool, default=False)
        parser.add_argument("--wandblog", type=str, default="")
        parser.add_argument("--project", type=str)

    def _add_subcommands(self, parser, **kwargs) -> None:
        """Adds subcommands to the input parser."""
        self._subcommand_parsers = {}
        parser_subcommands = parser.add_subcommands()
        # the user might have passed a builder function
        trainer_class = (
            self.trainer_class
            if isinstance(self.trainer_class, type)
            else class_from_function(self.trainer_class)
        )
        # register all subcommands in separate subcommand parsers under the main parser
        for subcommand in self.subcommands():
            fn = getattr(trainer_class, subcommand)
            # extract the first line description in the docstring for the subcommand help message
            description = _get_short_description(fn)
            subparser_kwargs = kwargs.get(subcommand, {})
            subparser_kwargs.setdefault("description", description)
            subcommand_parser = self._prepare_subcommand_parser(
                trainer_class, subcommand, **subparser_kwargs
            )
            self._subcommand_parsers[subcommand] = subcommand_parser
            parser_subcommands.add_subcommand(
                subcommand, subcommand_parser, help=description
            )
        description = (
            "setup everything for scprint2. only the default behavior is supported."
        )
        parser = self.init_parser(description=description)
        self._subcommand_parsers["easy_setup"] = parser
        parser_subcommands.add_subcommand("easy_setup", parser, help=description)
        for subcommand in TASKS:
            fn = getattr(subcommand[1], "__init__")
            description = _get_short_description(fn)
            subparser_kwargs = {}
            subparser_kwargs.setdefault("description", description)
            parser = self.init_parser(**subparser_kwargs)
            parser.add_argument(
                "--ckpt_path",
                type=str,
                help=("Path to the checkpoint to load."),
                required=True,
            )
            parser.add_argument(
                "--output_filename",
                type=str,
                help=("Path to the output file(s)."),
                required=True,
            )
            parser.add_argument(
                "--adata", type=str, help=("Path to the anndata file."), required=True
            )
            if subcommand[0] == "gninfer":
                parser.add_argument(
                    "--cell_type",
                    type=str,
                    help=("The cell type to infer the gene regulatory network."),
                    required=True,
                )
            parser.add_argument("--seed_everything", type=int, default=42)
            parser.add_argument(
                "--species",
                type=str,
                help=(
                    "If not included in the anndata under 'organism_ontology_term_id', the species of the dataset."
                ),
                required=False,
            )
            parser.add_argument("--use_raw", type=bool, default=False)
            parser.add_argument("--skip_validate", type=bool, default=True)
            parser.add_argument("--is_symbol", type=bool, default=False)
            parser.add_class_arguments(subcommand[1])
            added = parser.add_method_arguments(
                subcommand[1],
                "__call__",
                skip=set(["model", "adata", "cell_type"]),
            )

            self._subcommand_method_arguments[subcommand] = added
            self._subcommand_parsers[subcommand[0]] = parser
            parser_subcommands.add_subcommand(subcommand[0], parser, help=description)

    def _run_subcommand(self, subcommand: str) -> None:
        """Run the chosen subcommand."""
        before_fn = getattr(self, f"before_{subcommand}", None)
        if callable(before_fn):
            before_fn()
        if subcommand in self.subcommands():
            klass = self.trainer
            default = getattr(klass, subcommand)
            fn = getattr(self, subcommand, default)
            fn_kwargs = self._prepare_subcommand_kwargs(subcommand)
            fn(**fn_kwargs)
            after_fn = getattr(self, f"after_{subcommand}", None)
            if callable(after_fn):
                after_fn()
        elif subcommand == "easy_setup":
            from huggingface_hub import hf_hub_download
            from scdataloader.utils import (
                _adding_scbasecamp_genes,
                populate_my_ontology,
            )

            populate_my_ontology()
            _adding_scbasecamp_genes()
            hf_hub_download(
                repo_id="jkobject/scPRINT", filename="small-v2.ckpt", local_dir="."
            )
            print("finished command")
            return
        else:
            import numpy as np
            import scanpy as sc
            from scdataloader import Preprocessor
            from scdataloader.utils import load_genes

            from scprint2 import scPRINT2

            model_checkpoint_file = self.config_init[subcommand]["ckpt_path"]
            model = scPRINT2.load_from_checkpoint(
                model_checkpoint_file, precpt_gene_emb=None, gene_pos_file=None
            )

            adata = sc.read_h5ad(self.config_init[subcommand]["adata"])
            adata.obs.drop(columns="is_primary_data", inplace=True, errors="ignore")
            if self.config_init[subcommand]["species"] is not None:
                adata.obs["organism_ontology_term_id"] = self.config_init[subcommand][
                    "species"
                ]
            preprocessor = Preprocessor(
                do_postp=model.expr_emb_style == "metacell",
                force_preprocess=True,
                skip_validate=self.config_init[subcommand].get("skip_validate", True),
                use_raw=self.config_init[subcommand].get("use_raw", False),
                is_symbol=self.config_init[subcommand].get("is_symbol", False),
            )
            adata = preprocessor(adata)
            if model.expr_emb_style == "metacell":
                sc.pp.neighbors(adata, use_rep="X_pca")
            conf = dict(self.config_init[subcommand])
            if model.cell_transformer is None:
                if conf.get("ft_mode", "xpressor"):
                    print(
                        "model doesn't have xpressor, setting finetuning mode to 'full' model"
                    )
                conf["ft_mode"] = "full"
            missing = set(model.genes) - set(load_genes(model.organisms).index)
            if len(missing) > 0:
                print(
                    "Warning: some genes missmatch exist between model and ontology: solving...",
                )
                model._rm_genes(missing)
            if not torch.cuda.is_available():
                model = model.to(torch.float32)

            model = model.to("cuda" if torch.cuda.is_available() else "cpu")
            for key in [
                "seed_everything",
                "config",
                "species",
                "cache",
                "ckpt_path",
                "adata",
                "output_filename",
                "cell_type",
                "use_raw",
                "is_symbol",
                "skip_validate",
                "train_data",
                "val_data",
            ]:
                conf.pop(key, None)

            if subcommand == "embed":
                emb = Embedder(**conf)
                print("embedding...")
                adata, metrics = emb(
                    model=model,
                    adata=adata,
                )
                # save the anndata
                print("metrics:")
                print(metrics)
                print()
                name = (
                    self.config_init[subcommand]["output_filename"] + "_embedded.h5ad"
                )
                print(
                    "saving the file under the path: ",
                    name,
                )
                adata.write(name)
            elif subcommand == "gninfer":
                gn = GNInfer(**conf)
                adata = gn(
                    model=model,
                    adata=adata,
                    cell_type=self.config_init[subcommand]["cell_type"],
                )
                name = (
                    self.config_init[subcommand]["output_filename"]
                    + "_"
                    + self.config_init[subcommand]["cell_type"]
                    + "_grn.h5ad"
                )
                print(
                    "saving the file under the path: ",
                    name,
                )
                adata.write(name)

            elif subcommand == "denoise":
                dn = Denoiser(**conf)
                metrics, _, expr_pred = dn(
                    model=model,
                    adata=adata,
                )
                print("metrics:")
                print(metrics)
                print()
                # now what we are doing here it to complete the expression profile with the denoised values.
                # this is not done by default for now
                name = (
                    self.config_init[subcommand]["output_filename"] + "_denoised.h5ad"
                )
                print(
                    "saving the file under the path: ",
                    name,
                )
                expr_pred.write(name)
            elif subcommand == "finetune":
                ft = FinetuneBatchClass(**conf)
                model = ft(
                    model=model,
                    adata=adata,
                )
                name = (
                    self.config_init[subcommand]["output_filename"] + "_finetuned.ckpt"
                )
                print(
                    "saving the model under the path: ",
                    name,
                )
                model.save_checkpoint(name)
            elif subcommand == "gene_emb":
                ge = GeneEmbeddingExtractor(**conf)
                ada = ge(
                    model=model,
                    adata=adata,
                )
                name = (
                    self.config_init[subcommand]["output_filename"] + "_gene_emb.h5ad"
                )
                print(
                    "saving the file under the path: ",
                    name,
                )
                ada.write(name)
            elif subcommand == "impute":
                imp = Imputer(**conf)
                _, expr_pred = imp(
                    model=model,
                    adata=adata,
                )
                name = self.config_init[subcommand]["output_filename"] + "_imputed.h5ad"
                print(
                    "saving the file under the path: ",
                    name,
                )
                expr_pred.write(name)
            else:
                raise ValueError(f"Unknown subcommand: {subcommand}")

    def before_instantiate_classes(self):
        for k, v in self.config.items():
            if "set_float32_matmul_precision" in k:
                if v:
                    torch.set_float32_matmul_precision("medium")
        import os

        simpler_f_hash = os.system("cd ../simpler_flash && git rev-parse HEAD")
        scdata_v = os.system("cd ../scdataloader && git rev-parse HEAD")
        print("scdata_v:", scdata_v)
        print("simpler_f_hash:", simpler_f_hash)
        if "fit" in self.config and self.config["fit"]["trainer"]["strategy"] in [
            "ddp",
            "ddp_find_unused_parameters_true",
            "fsdp",
        ]:
            import os

            os.environ["NCCL_TIMEOUT"] = str(TIMEOUT)  # 9 hours in seconds
            os.environ["TORCH_DISTRIBUTED_TIMEOUT"] = str(TIMEOUT)  # 9 hours in seconds
            os.environ["PL_TRAINER_STRATEGY_TIMEOUT"] = str(TIMEOUT)

            print(f"setting global pytorch distributed timeout to {TIMEOUT}s")

    def after_instantiate_classes(self):
        try:
            self.model.name = self.trainer._loggers[0].version
        except:
            print("not on wandb, could not set name")
        if hasattr(self, "datamodule") and self.datamodule is not None:
            self.datamodule.set_valid_genes_collator(self.model.genes)

    def instantiate_trainer(self, **kwargs) -> Trainer:
        """Override to customize trainer instantiation"""
        # Modify strategy if it's DDP
        trainer = super().instantiate_trainer(**kwargs)
        if "fit" in self.config and self.config["fit"]["trainer"]["strategy"] in [
            "ddp",
            "ddp_find_unused_parameters_true",
            "fsdp",
        ]:
            # Create DDPStrategy with custom timeout
            from datetime import timedelta

            # Update the config
            print("updating the config")
            trainer.strategy._timeout = timedelta(seconds=TIMEOUT)  # 9 hours in seconds
            trainer.strategy.setup_distributed()
        # Call parent method to create trainer
        return trainer



import gc
import math
import os.path
from collections import Counter
from typing import Dict, List, Optional, Union

import bionty as bt
import numpy as np
import pandas as pd
import scanpy as sc
import torch
from anndata import AnnData
from matplotlib import pyplot as plt
from scdataloader.utils import translate
from scipy.sparse import csr_matrix
from torch import Tensor
from torch.distributions import Gamma, Poisson

from .. import utils
from ..datasets import DENOISE_DATASETS, EMBEDDING_DATASETS
from ..tasks import cell_emb as embbed_task
from ..tasks import denoise as denoise_task
from ..tasks import grn as grn_task

FILEDIR = os.path.dirname(os.path.realpath(__file__))


def make_adata(
    genes: List[str],
    embs: Union[Tensor, Dict[str, Tensor]],
    pos: Tensor = None,
    expr_pred: List[Tensor] = None,
    classes: List[str] = None,
    pred: Tensor = None,
    label_decoders: Optional[Dict] = None,
    labels_hierarchy: Optional[Dict] = None,
    gtclass: Optional[Tensor] = None,
    doplot: bool = True,
) -> AnnData:
    """
    This function creates an AnnData object from the given input parameters.

    Args:
        genes (list): List of genes that will be used as variable names.
        embs (torch.Tensor|Dict): Embeddings of the cells. The shape of the tensor is (n_cells, n_features).
            if multiple, it is a dict of name -> tensor
        pos (torch.Tensor): Positions of the cells. The shape of the tensor is (n_cells,).
        expr_pred (List[torch.Tensor]): Predicted expression. The shape of the tensors are (n_cells, n_genes).
            the first is mu, the second theta, the third pi if present
        classes (list): List of classes, the order should be the same as in the pred and gtclass tensors.
        pred (torch.Tensor, optional): Predicted labels. The shape of the tensor is (n_cells, n_classes). Default is None.
        label_decoders (dict, optional): Dictionary to map class codes to class names. Default is None.
        labels_hierarchy (dict, optional): Dictionary representing the hierarchy of labels. Default is {}. see the model for defintion.
        gtclass (torch.Tensor, optional): Ground truth class values. Default is None.
        doplot (bool, optional): Whether to generate plots. Default is True.

    Returns:
        anndata.AnnData: The created AnnData object.
    """
    print("logging the anndata")
    colname = ["pred_" + i for i in classes]
    if pred is not None:
        obs = np.array(pred.to(device="cpu", dtype=torch.int32))
        # label decoders is not cls_decoders. one is a dict to map class codes (ints)
        # to class names the other is the module the predict the class
        if label_decoders is not None:
            obs = np.array(
                [
                    [label_decoders[classes[i]][n] for n in name]
                    for i, name in enumerate(obs.T)
                ]
            ).T
        if gtclass is not None:
            colname += classes
            nobs = np.array(gtclass.to(device="cpu", dtype=torch.int32))
            if label_decoders is not None:
                nobs = np.array(
                    [
                        [label_decoders[classes[i]][n] for n in name]
                        for i, name in enumerate(nobs.T)
                    ]
                ).T
            obs = np.hstack([obs, nobs])

    n_cells = embs[list(embs.keys())[0]].shape[0]
    layers = None
    size = len(genes)
    if pos is not None:
        minval = pos.min()
        maxval = pos.max()
        genes = genes[minval : maxval + 1]
        size = len(genes)
        pos = pos - minval
        mu_array = np.zeros((n_cells, size), dtype=np.float32)
        pos = pos.cpu().numpy()
        # Create empty array with same shape as expr_pred[0]
        # Fill array with values from expr_pred[0]
        for idx in range(n_cells):
            mu_array[idx, pos[idx]] = expr_pred[0][idx].cpu().numpy() + 1
        exist = mu_array.sum(0) != 0
        mu_array = mu_array[:, exist]
        mu_array[mu_array == 1] = 0
        layers = {
            "scprint_mu": mu_array,
            #  "used_scprint": csr_matrix(pos),
        }
        if len(expr_pred) > 1:
            theta_array = np.zeros((n_cells, size), dtype=np.float32)
            # Fill array with values from expr_pred[0]
            for idx in range(n_cells):
                theta_array[idx, pos[idx]] = expr_pred[1][idx].cpu().numpy()
            layers["scprint_theta"] = theta_array[:, exist]

            pi_array = np.zeros((n_cells, size), dtype=np.float32)
            # Fill array with values from expr_pred[0]
            for idx in range(n_cells):
                pi_array[idx, pos[idx]] = expr_pred[2][idx].cpu().numpy()
            layers["scprint_pi"] = pi_array[:, exist]
        genes = [n for i, n in enumerate(genes) if exist[i] > 0]
    else:
        genes = []
    adata = AnnData(
        X=csr_matrix((n_cells, len(genes))),
        layers=layers,
        obs=(
            pd.DataFrame(
                obs,
                columns=colname,
            )
            if pred is not None
            else None
        ),
        var=pd.DataFrame(index=genes),
    )

    for k, v in embs.items():
        adata.obsm["scprint_emb_" + k] = v.cpu().numpy()
        rep = "scprint_emb_" + k
    del embs
    accuracy = {}
    if labels_hierarchy is None:
        labels_hierarchy = {}
    if pred is not None:
        for clss in classes:
            if gtclass is not None:
                tr = translate(set(adata.obs[clss]), clss)
                if tr is not None:
                    adata.obs["conv_" + clss] = adata.obs[clss].replace(tr)
            tr = translate(set(adata.obs["pred_" + clss]), clss)
            if tr is not None:
                adata.obs["conv_pred_" + clss] = adata.obs["pred_" + clss].replace(tr)
            res = []
            if label_decoders is not None and gtclass is not None:
                class_topred = label_decoders[clss].values()
                if clss in labels_hierarchy:
                    cur_labels_hierarchy = {
                        label_decoders[clss][k]: [label_decoders[clss][i] for i in v]
                        for k, v in labels_hierarchy[clss].items()
                    }
                else:
                    cur_labels_hierarchy = {}
                for pred, true in adata.obs[["pred_" + clss, clss]].values:
                    if pred == true:
                        res.append(True)
                        continue
                    if len(labels_hierarchy) > 0:
                        if true in cur_labels_hierarchy:
                            res.append(pred in cur_labels_hierarchy[true])
                        elif true not in class_topred:
                            raise ValueError(
                                f"true label {true} not in available classes"
                            )
                        elif true != "unknown":
                            res.append(False)
                    elif true not in class_topred:
                        raise ValueError(f"true label {true} not in available classes")
                    elif true != "unknown":
                        res.append(False)
                    else:
                        pass
                accuracy["pred_" + clss] = sum(res) / len(res) if len(res) > 0 else 0
        adata.obs = adata.obs.astype("category")
    print(adata)
    if doplot and adata.shape[0] > 100:
        sc.pp.neighbors(adata, use_rep=rep)
        sc.tl.umap(adata)
        sc.tl.leiden(adata, key_added="sprint_leiden")
        if gtclass is not None:
            color = [
                i
                for pair in zip(
                    [
                        "conv_" + i if "conv_" + i in adata.obs.columns else i
                        for i in classes
                    ],
                    [
                        (
                            "conv_pred_" + i
                            if "conv_pred_" + i in adata.obs.columns
                            else "pred_" + i
                        )
                        for i in classes
                    ],
                )
                for i in pair
            ]
            fig, axs = plt.subplots(
                int(len(color) / 2), 2, figsize=(24, len(color) * 4)
            )
            plt.subplots_adjust(wspace=1)
            if len(color) > 2:
                for i, col in enumerate(color):
                    sc.pl.umap(
                        adata,
                        color=col,
                        ax=axs[i // 2, i % 2],
                        show=False,
                    )
                    acc = ""
                    if "pred_" in col and col.split("conv_")[-1] in accuracy:
                        acc = " (accuracy: {:.2f})".format(
                            accuracy[col.split("conv_")[-1]]
                        )
                    axs[i // 2, i % 2].set_title(col + " UMAP" + acc)
                    if "cell_type" in col:
                        axs[i // 2, i % 2].legend(fontsize="x-small")
                    axs[i // 2, i % 2].set_xlabel("UMAP1")
                    axs[i // 2, i % 2].set_ylabel("UMAP2")
            else:
                for i, col in enumerate(color):
                    sc.pl.umap(
                        adata,
                        color=col,
                        ax=axs[i % 2],
                        show=False,
                    )
                    acc = ""
                    if "pred_" in col and col.split("conv_")[-1] in accuracy:
                        acc = " (accuracy: {:.2f})".format(
                            accuracy[col.split("conv_")[-1]]
                        )
                    axs[i % 2].set_title(col + " UMAP" + acc)
                    if "cell_type" in col:
                        axs[i % 2].legend(fontsize="x-small")
                    axs[i % 2].set_xlabel("UMAP1")
                    axs[i % 2].set_ylabel("UMAP2")
        else:
            color = [
                (
                    "conv_pred_" + i
                    if "conv_pred_" + i in adata.obs.columns
                    else "pred_" + i
                )
                for i in classes
            ]
            if len(color) > 1:
                fig, axs = plt.subplots(len(color), 1, figsize=(16, len(color) * 8))
                for i, col in enumerate(color):
                    sc.pl.umap(
                        adata,
                        color=col,
                        ax=axs[i],
                        show=False,
                    )
                    acc = ""
                    if "pred_" in col and col.split("conv_")[-1] in accuracy:
                        acc = " (accuracy: {:.2f})".format(
                            accuracy[col.split("conv_")[-1]]
                        )
                    axs[i].set_title(col + " UMAP of " + rep + " embedding " + acc)
                    axs[i].set_xlabel("UMAP1")
                    axs[i].set_ylabel("UMAP2")
            else:
                fig = sc.pl.umap(
                    adata,
                    color=color,
                    show=False,
                    return_fig=True,
                )
        plt.show()
    else:
        fig = None
    return adata, fig


def _init_weights(
    module: torch.nn.Module,
    n_layer: int,
    initializer_range: float = 0.02,
    mup_width_scale: float = 1.0,
    rescale_prenorm_residual: bool = True,
):
    """
    This function initializes the weights of the given module. The initialization is done based on the type of the module.

    If the module is a Linear layer, the weights are initialized with a normal distribution with a standard deviation
    that is a product of the initializer range and the mup_init_scale. The learning rate multiplier is also set for the
    weights of the Linear layer. If the module has a bias, it is initialized with zeros.
    If the module is an Embedding layer, no initialization is performed.
    If the rescale_prenorm_residual flag is set to True, the weights of the residual layers are reinitialized according
    to the scheme proposed in the OpenAI GPT-2 Paper. The weights are scaled by a factor of 1/sqrt(N), where N is the
    number of residual layers.

    Args:
        module (torch.nn.Module): The module whose weights are to be initialized.
        n_layer (int): The number of layers in the module.
        initializer_range (float, optional): The range of the initializer. Defaults to 0.02.
        mup_width_scale (float, optional): The scale for the mup initialization. Defaults to 1.0.
        rescale_prenorm_residual (bool, optional): Flag to indicate whether to rescale the prenorm residual. Defaults to True.
    """
    mup_init_scale = math.sqrt(mup_width_scale)
    if isinstance(module, torch.nn.Linear):
        torch.nn.init.normal_(module.weight, std=initializer_range * mup_init_scale)
        optim_cfg = getattr(module.weight, "_optim", {})
        optim_cfg.update({"lr_multiplier": mup_width_scale})
        setattr(module.weight, "_optim", optim_cfg)
        if module.bias is not None:
            torch.nn.init.zeros_(module.bias)
    elif isinstance(module, torch.nn.Embedding):
        pass

    if rescale_prenorm_residual:
        # Reinitialize selected weights subject to the OpenAI GPT-2 Paper Scheme:
        #   > A modified initialization which accounts for the accumulation on the residual path with model depth. Scale
        #   > the weights of residual layers at initialization by a factor of 1/√N where N is the # of residual layers.
        #   >   -- GPT-2 :: https://openai.com/blog/better-language-models/
        #
        # Reference (Megatron-LM): https://github.com/NVIDIA/Megatron-LM/blob/main/megatron/model/gpt_model.py
        for name, p in module.named_parameters():
            if name in ["out_proj.weight", "fc2.weight"]:
                # Special Scaled Initialization --> There are 2 Layer Norms per Transformer Block
                torch.nn.init.normal_(
                    p,
                    mean=0.0,
                    std=initializer_range * mup_init_scale / math.sqrt(2 * n_layer),
                )


def downsample_profile(mat: Tensor, dropout: float, method="new", randsamp=False) -> Tensor:
    """
    This function downsamples the expression profile of a given single cell RNA matrix.

    The noise is applied based on the renoise parameter,
    the total counts of the matrix, and the number of genes. The function first calculates the noise
    threshold (scaler) based on the renoise parameter. It then generates an initial matrix count by
    applying a Poisson distribution to a random tensor scaled by the total counts and the number of genes.
    The function then models the sampling zeros by applying a Poisson distribution to a random tensor
    scaled by the noise threshold, the total counts, and the number of genes. The function also models
    the technical zeros by generating a random tensor and comparing it to the noise threshold. The final
    matrix count is calculated by subtracting the sampling zeros from the initial matrix count and
    multiplying by the technical zeros. The function ensures that the final matrix count is not less
    than zero by taking the maximum of the final matrix count and a tensor of zeros. The function
    returns the final matrix count.

    Args:
        mat (torch.Tensor): The input matrix.
        dropout (float): The renoise parameter.

    Returns:
        torch.Tensor: The matrix count after applying noise.
    """
    # Randomly drop on average N counts to each element of expression using a heavy tail Gaussian distribution
    # here we try to get the scale of the distribution so as to remove the right number of counts from each gene
    # https://genomebiology.biomedcentral.com/articles/10.1186/s13059-022-02601-5#:~:text=Zero%20measurements%20in%20scRNA%2Dseq,generation%20of%20scRNA%2Dseq%20data.
    if randsamp:
        dropout = torch.rand(mat.shape[0], device=mat.device) * dropout
        dropout = (
            dropout.unsqueeze(1)
            if len(mat.shape) == 2
            else dropout.unsqueeze(1).unsqueeze(1)
        )
    if method == "old":
        totcounts = mat.sum(-1)
        ngenes = mat.shape[-1]
        tnoise = 1 - (1 - dropout) ** (1 / 2)
        # we model the sampling zeros (dropping 30% of the reads)
        res = torch.poisson(
            torch.rand(mat.shape, device=mat.device)
            * ((tnoise * totcounts.unsqueeze(-1)) / (0.5 * ngenes))
        ).int()
        # we model the technical zeros (dropping 50% of the genes)
        drop = (torch.rand(mat.shape, device=mat.device) > tnoise).int()

        mat = (mat - res) * drop
        return torch.maximum(
            mat,
            torch.zeros(
                (1, 1) if len(mat.shape) == 2 else (1, 1, 1),
                device=mat.device,
                dtype=torch.int,
            ),
        )
    elif method == "jules":
        scaler = (1 - dropout) ** (1 / 2)
        notdrop = (
            torch.rand(
                mat.shape,
                device=mat.device,
            )
            < scaler
        ).int()
        notdrop[mat == 0] = 0
        # apply the dropout after the poisson, right?
        return notdrop * torch.poisson(mat * scaler)
    elif method == "new":
        dropout = dropout * 1.1
        # we model the sampling zeros (dropping 30% of the reads)
        res = torch.poisson((mat * (dropout / 2))).int()
        # we model the technical zeros (dropping 50% of the genes)
        notdrop = (torch.rand(mat.shape, device=mat.device) >= (dropout / 2)).int()
        mat = (mat - res) * notdrop
        return torch.maximum(
            mat,
            torch.zeros(
                (1, 1) if len(mat.shape) == 2 else (1, 1, 1),
                device=mat.device,
                dtype=torch.int,
            ),
        )
    else:
        raise ValueError(f"method {method} not recognized")


def simple_masker(
    shape: List[int],
    mask_ratio: float = 0.15,
) -> torch.Tensor:
    """
    Randomly mask a batch of data.

    Args:
        shape (List[int]): The shape of the data.
        mask_ratio (float): The ratio of genes to mask, default to 0.15.

    Returns:
        torch.Tensor: A tensor of masked data.
    """
    return torch.rand(shape) < mask_ratio


class WeightedMasker:
    def __init__(
        self,
        genes: List[str],
        TFs: List[str] = utils.fileToList(FILEDIR + "/../../data/main/TFs.txt"),
        tf_weight: float = 10,
    ):
        """
        Randomly mask a batch of data.

        Args:
            genes (List[str]): The list of genes the model might see.
            TFs (List[str]): The list of TFs the model can drop.
            tf_weight (float): How likely it is to drop a non TF compared to a TF.
        """
        TFs = set(TFs)
        self.weights = torch.tensor([tf_weight if gene in TFs else 1 for gene in genes])
        self.max_to_drop = (self.weights == tf_weight).sum()
        self.tf_weight = tf_weight

    def __call__(self, ids: torch.Tensor, mask_ratio: float = 1.0) -> torch.Tensor:
        if self.tf_weight == 0:
            if mask_ratio * ids.shape[1] > self.max_to_drop:
                raise ValueError("Cannot drop more than max_to_drop")
        # Create a tensor of probabilities for each position
        probs = self.weights.expand(ids.shape[0], -1).to(ids.device)
        ids = ids.to(torch.int64)
        probs = torch.gather(
            probs, 1, ids
        )  # Get probabilities only for the indices in ids
        probs = probs / probs.sum(1, keepdim=True)

        # Sample from multinomial for each item in batch
        num_samples = int(ids.shape[1] * mask_ratio)
        mask = torch.zeros_like(ids, dtype=torch.bool)
        sampled = torch.multinomial(probs, num_samples, replacement=False)
        return mask.scatter_(1, sampled, True)


def zinb_sample(
    mu: torch.Tensor,
    theta: torch.Tensor,
    zi_probs: torch.Tensor,
    sample_shape: torch.Size = torch.Size([]),
) -> torch.Tensor:
    """
    zinb_sample This function generates a sample from a Zero-Inflated Negative Binomial (ZINB) distribution.

    Args:
        mu (torch.Tensor): The mean of the Negative Binomial (NB) distribution.
        theta (torch.Tensor): The dispersion parameter of the NB distribution.
        zi_probs (torch.Tensor): The zero-inflation probabilities.
        sample_shape (torch.Size, optional): The output shape. Defaults to torch.Size([]).

    Returns:
        torch.Tensor: A sample from the ZINB distribution.
    """
    concentration = theta
    rate = theta / mu
    # Important remark: Gamma is parametrized by the rate = 1/scale!
    gamma_d = Gamma(concentration=concentration, rate=rate)
    p_means = gamma_d.sample(sample_shape)

    # Clamping as distributions objects can have buggy behaviors when
    # their parameters are too high
    l_train = torch.clamp(p_means, max=1e8)
    samp = Poisson(l_train).sample()  # Shape : (n_samples, n_cells_batch, n_vars)
    is_zero = torch.rand_like(samp) <= zi_probs
    samp_ = torch.where(is_zero, torch.zeros_like(samp), samp)
    return samp_


class Attention:
    def __init__(
        self,
        gene_dim: int,
        precomp_attn: bool = False,
        apply_softmax: bool = True,
        sum_heads: bool = True,
        additional_tokens: int = 0,
    ):
        """
        Initialize the Attention class.

        Args:
            gene_dim (int): The dimension of the gene.
            additional_tokens (int): The number of additional tokens to add.
            precomp_attn (bool): Whether to compute attention or it is precomputed
            apply_softmax (bool): Whether to apply softmax to the attention.
            sum_heads (bool): Whether to sum the heads.
        """
        self.data: Optional[Tensor] = None
        self.gene_dim: int = gene_dim
        self.additional_tokens: int = additional_tokens
        self.div: Optional[Tensor] = None
        self.apply_softmax: bool = apply_softmax
        self.sum_heads: bool = sum_heads
        self.precomp_attn: bool = precomp_attn
        self.speciesloc: int = 0

    def add(self, *args, **kwargs) -> None:
        if not self.precomp_attn:
            self.add_qk(*args, **kwargs)
        else:
            self.add_attn(*args, **kwargs)

    def add_attn(
        self, x: List[Tensor], pos: Tensor, expr: Optional[Tensor] = None
    ) -> None:
        """
        Aggregate the attention or data based on the precomp_attn flag.

        Args:
            x (List[Tensor]): List of tensors to aggregate. Tensor of size (batch, seq_len, 2, heads, emb)
            pos (Tensor): Position tensor.
        """
        if self.data is None:
            self.data = torch.zeros(
                [
                    self.gene_dim + self.additional_tokens,
                    self.gene_dim + self.additional_tokens,
                ],
                device=pos.device,
                dtype=torch.float32,
            )
            self.div = torch.zeros(1, device=pos.device, dtype=torch.float32)
        for i, elem in enumerate(x):
            if self.apply_softmax:
                attn = torch.nn.functional.softmax(
                    elem[:, :, 0, :, :].permute(0, 2, 1, 3)
                    @ elem[:, :, 1, :, :].permute(0, 2, 3, 1),
                    dim=-1,
                )
                if expr is not None:
                    attn[:, :, self.additional_tokens :, self.additional_tokens :] = (
                        attn[:, :, self.additional_tokens :, self.additional_tokens :]
                        * (expr > 0).float().unsqueeze(1).unsqueeze(-1)
                        * (expr > 0).float().unsqueeze(1).unsqueeze(2)
                    )
                self.data += attn.sum(0).mean(0)
            else:
                self.data[:, :] += (
                    (
                        elem[:, :, 0, :, :].permute(0, 2, 1, 3)
                        @ elem[:, :, 1, :, :].permute(0, 2, 3, 1)
                    )
                    .sum(0)
                    .mean(0)
                )
            self.div += 1

    def add_qk(
        self, x: List[Tensor], pos: Tensor, expr: Optional[Tensor] = None
    ) -> None:
        """
        Add data to the internal storage.

        Args:
            x (List[Tensor]): List of tensors to add.
            pos (Tensor): Position tensor.
        """
        # this is a debugger line
        if self.data is None:
            self.data = torch.zeros(
                [len(x), self.gene_dim + self.additional_tokens] + list(x[0].shape[2:]),
                device=pos.device,
            )
            self.div = torch.zeros(
                self.gene_dim + self.additional_tokens, device=pos.device
            )
        for i in range(x[0].shape[0]):  # batch size
            loc = torch.cat(
                [
                    torch.arange(self.additional_tokens, device=pos.device),
                    pos[i] + self.additional_tokens - self.speciesloc,
                ]
            ).int()
            for j in range(len(x)):  # number of layers * heads
                self.data[j, loc, :, :, :] += x[j][i]
            self.div[loc] += 1

    def get(self) -> Optional[np.ndarray]:
        """
        Get the aggregated attention or data.

        Returns:
            Optional[np.ndarray]: The aggregated attention or data.
        """
        if not self.precomp_attn:
            if self.data is None:
                return None
            # shape is (layers, genes, qkv, heads, emb)
            return self.data / self.div.view(1, self.div.shape[0], 1, 1, 1)
        else:
            if self.data is None:
                return None
            self.data.div_(self.div)
            return self.data


def test(
    model: torch.nn.Module,
    filedir: str,
    do_class: bool = True,
    maxcells_grn: int = 1024,
) -> None:
    """
    Test the given model on the full set of benchmarks and save the results to JSON files.

    Args:
        model (torch.nn.Module): The model to be tested.
        filedir (str): The directory where the data files are located.
        do_class (bool): Whether to perform classification. Defaults to True.
        maxcells_grn (int): Maximum cells for GRN analysis. Defaults to 1024.

    Returns:
        None
    """
    metrics = {}
    tot = {}
    for dataset, path in EMBEDDING_DATASETS.items():
        res = embbed_task.default_benchmark(
            model,
            dataset=path,
            do_class=do_class,
            coarse=False,
        )
        tot["embed_" + dataset] = res
        metrics.update(
            {
                "emb_" + dataset + "/scib": float(res["scib"]["Total"]),
                "emb_" + dataset + "/scib_bio": float(res["scib"]["Bio conservation"]),
                "emb_"
                + dataset
                + "/scib_batch": float(res["scib"]["Batch correction"]),
                "emb_"
                + dataset
                + "/ct_class": float(
                    res["classif"].get("cell_type_ontology_term_id", {}).get("macro", 0)
                    if do_class
                    else 0
                ),
                "emb_"
                + dataset
                + "/ct_class_macro": float(
                    res["classif"].get("cell_type_ontology_term_id", {}).get("macro", 0)
                    if do_class
                    else 0
                ),
            }
        )
        print(metrics)
        gc.collect()
    for dataset, filepath in DENOISE_DATASETS.items():
        res = denoise_task.default_benchmark(model, dataset=filepath)
        tot["denoise_" + dataset] = res
        metrics.update(
            {
                "denoise_"
                + dataset
                + "/reco2full_vs_noisy2full": float(
                    res["reco2full"] - res["noisy2full"]
                ),
            }
        )
        print(metrics)
        gc.collect()
    res = grn_task.default_benchmark(
        model,
        "gwps",
        batch_size=32 if model.d_model <= 512 else 8,
        maxcells=maxcells_grn,
    )
    tot["grn_gwps"] = res
    metrics.update(
        {
            "grn_gwps/auprc_self": float(res["self"]["auprc"]),
            "grn_gwps/epr_self": float(res["self"]["epr"]),
            "grn_gwps/auprc_omni": float(res["omni"]["auprc"]),
            "grn_gwps/epr_omni": float(res["omni"]["epr"]),
            "grn_gwps/auprc": float(res["mean"]["auprc"]),
            "grn_gwps/epr": float(res["mean"]["epr"]),
        }
    )
    print(metrics)
    gc.collect()
    for dataset, filepath in {
        "old_kidney": "https://datasets.cellxgene.cziscience.com/ede85b09-454b-4374-bf60-5f675e989b64.h5ad",
        # "kidney": "https://datasets.cellxgene.cziscience.com/01bc7039-961f-4c24-b407-d535a2a7ba2c.h5ad",
        "lung_smart": "https://datasets.cellxgene.cziscience.com/6ebba0e0-a159-406f-8095-451115673a2c.h5ad",
        # filedir + "/../../data/yBCKp6HmXuHa0cZptMo7.h5ad",
    }.items():
        res = grn_task.default_benchmark(
            model,
            filepath,
            # kidney dataset (2.87, 1.27) (0.00147, 0.00133)
            batch_size=32 if model.d_model <= 512 else 8,
            maxcells=maxcells_grn,
            maxgenes=4000,
        )
        tot["grn_omni_" + dataset] = res
        metrics.update(
            {
                "grn_omni_"
                + dataset
                + "/auprc_class": float(
                    np.mean([i["auprc"] for k, i in res.items() if "_class" in k])
                ),
                "grn_omni_"
                + dataset
                + "/or_class": float(
                    np.mean([i["odd_ratio"] for k, i in res.items() if "_class" in k])
                ),
                "grn_omni_"
                + dataset
                + "/tf_enr_class": float(
                    np.sum(
                        [
                            i.get("TF_enr", False)
                            for k, i in res.items()
                            if "_class" in k
                        ]
                    )
                ),
                "grn_omni_"
                + dataset
                + "/tf_targ_enr_class": float(
                    np.mean(
                        [
                            i["significant_enriched_TFtargets"]
                            for k, i in res.items()
                            if "_class" in k
                        ]
                    )
                ),
                "grn_omni_"
                + dataset
                + "/auprc": float(
                    np.mean([i["auprc"] for k, i in res.items() if "_mean" in k])
                ),
                "grn_omni_"
                + dataset
                + "/epr": float(
                    np.mean([i["epr"] for k, i in res.items() if "_mean" in k])
                ),
                "grn_omni_"
                + dataset
                + "/or": float(
                    np.mean([i["odd_ratio"] for k, i in res.items() if "_mean" in k])
                ),
                "grn_omni_"
                + dataset
                + "/tf_enr": float(
                    np.sum(
                        [i.get("TF_enr", False) for k, i in res.items() if "_mean" in k]
                    )
                ),
                "grn_omni_"
                + dataset
                + "/tf_targ_enr": float(
                    np.mean(
                        [
                            i["significant_enriched_TFtargets"]
                            for k, i in res.items()
                            if "_mean" in k
                        ]
                    )
                ),
                # 'grn_omni/ct': res['classif']['cell_type_ontology_term_id']['accuracy'],
            }
        )
        print(metrics)
        gc.collect()
    return metrics, tot


def relabel_assay_for_adv(label_decoders, labels_hierarchy):
    topred = [
        "EFO:0700011",
        "EFO:0008953",
        "EFO:0030003",
        "EFO:0700016",
        "EFO:0700010",
        "EFO:0022490",
        "EFO:0008722",
        "EFO:0010961",
        "EFO:0008720",
        "EFO:0010550",
        "EFO:0030004",
        "EFO:0008931",
        "EFO:0010010",
        "EFO:0008780",
        "EFO:0009919",
        "EFO:0010713",
        "EFO:0030059",
        "EFO:0008796",
        "EFO:0030002",
        "EFO:0030062",
        "EFO:0008919",
        "EFO:0008930",
        "EFO:0010184",
        "EFO:0008995",
        "EFO:0030080",
        "EFO:0700003",
    ]
    """
    relabel_assay_for_adv a method that groups assays so that the adv classification is easier and doesn't overfit

    Returns:
        dict: a remapping dictionary for grouping the labels to their coarser names when appropriate
    """
    relab = {}
    enc = {v: k for k, v in label_decoders["assay_ontology_term_id"].items()}
    for i in enc.keys():
        if i in topred:
            relab[enc[i]] = topred.index(i)
        else:
            relab[enc[i]] = -1
            prevlen = 10000
            for val in topred:
                li = labels_hierarchy["assay_ontology_term_id"].get(
                    enc.get(val, ""), []
                )
                if enc[i] in li and prevlen > len(li):
                    relab[enc[i]] = topred.index(val)
                    prevlen = len(li)
    return relab



import os
from typing import Any, List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import scanpy as sc
import sklearn.metrics
import torch
from anndata import AnnData, concat
from scdataloader import Collator, Preprocessor
from scdataloader.data import SimpleAnnDataset
from scdataloader.utils import get_descendants, random_str
from scipy.stats import spearmanr
from simpler_flash import FlashTransformer
from torch.nn import functional as F
from torch.utils.data import DataLoader
from tqdm import tqdm

from scprint2.model import utils
from scprint2.tasks.denoise import plot_cell_depth_wise_corr_improvement

FILE_DIR = os.path.dirname(os.path.realpath(__file__))


class Imputer:
    def __init__(
        self,
        genes_to_use: List[str],
        genes_to_impute: List[str],
        batch_size: int = 10,
        num_workers: int = 1,
        max_cells: int = 500_000,
        doplot: bool = False,
        method: str = "generative",
        predict_depth_mult: int = 4,
        save_every: int = 100_000,
        apply_zero_pred: bool = True,
        use_knn: bool = True,
    ):
        """
        Imputer class for imputing missing values in scRNA-seq data using a scPRINT model

        Args:
            batch_size (int, optional): Batch size for processing. Defaults to 10.
            num_workers (int, optional): Number of workers for data loading. Defaults to 1.
            max_cells (int, optional): Number of cells to use for plotting correlation. Defaults to 10000.
            doplot (bool, optional): Whether to generate plots of the similarity between the denoised and true expression data. Defaults to False.
                Only works when downsample_expr is not None and max_cells < 100.
            method (str, optional): Imputation method, either 'masking' or 'generative'. Defaults to 'generative'.
            predict_depth_mult (int, optional): Multiplier for prediction depth. Defaults to 4.
                This will artificially increase the sequencing depth (or number of counts) to 4 times the original depth.
            genes_to_use (List[str]): List of genes to use for imputation. Defaults to None.
            genes_to_impute (List[str]): List of genes to impute. Defaults to None.
            save_every (int, optional): The number of cells to save at a time. Defaults to 100_000.
                This is important to avoid memory issues.
            apply_zero_pred (bool, optional): Whether to apply zero prediction adjustment. Defaults to True.
                applying zero inflation might give results closer to the specific biases of sequencing technologies but less biological truthful.
            use_knn (bool, optional): Whether to use k-nearest neighbors information. Defaults to True.
        """
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.max_cells = max_cells
        self.doplot = doplot
        self.predict_depth_mult = predict_depth_mult
        self.save_every = save_every
        self.genes_to_use = genes_to_use
        self.genes_to_impute = genes_to_impute
        self.method = method
        self.apply_zero_pred = apply_zero_pred
        self.use_knn = use_knn

    def __call__(
        self, model: torch.nn.Module, adata: AnnData
    ) -> tuple[Optional[np.ndarray], AnnData]:
        """
        __call__ calling the function

        Args:
            model (torch.nn.Module): The scPRINT model to be used for denoising.
            adata (AnnData): The anndata of shape n_obs x n_vars. Rows correspond to cells and columns to genes.

        Returns:
            Optional[np.ndarray]: The random indices of the cells used when max_cells < adata.shape[0].
            AnnData: The imputed anndata.
        """
        # Select random number
        random_indices = None
        if self.max_cells < adata.shape[0]:
            random_indices = np.random.randint(
                low=0, high=adata.shape[0], size=self.max_cells
            )
            adataset = SimpleAnnDataset(
                adata[random_indices],
                obs_to_output=["organism_ontology_term_id"],
                get_knn_cells=model.expr_emb_style == "metacell" and self.use_knn,
            )
        else:
            adataset = SimpleAnnDataset(
                adata,
                obs_to_output=["organism_ontology_term_id"],
                get_knn_cells=model.expr_emb_style == "metacell" and self.use_knn,
            )
        genes_to_use = set(model.genes) & set(self.genes_to_use)
        print(
            f"{100 * len(genes_to_use) / len(self.genes_to_use)}% of genes to use are available in the model"
        )
        genes_to_impute = set(model.genes) & set(self.genes_to_impute)
        print(
            f"{100 * len(genes_to_impute) / len(self.genes_to_impute)}% of genes to impute are available in the model"
        )
        tot = genes_to_use | genes_to_impute
        tot = sorted(tot)
        col = Collator(
            organisms=model.organisms,
            valid_genes=model.genes,
            how="some",
            genelist=list(genes_to_use)
            + (list(genes_to_impute) if self.method == "masking" else []),
            n_bins=model.n_input_bins if model.expr_emb_style == "binned" else 0,
        )
        dataloader = DataLoader(
            adataset,
            collate_fn=col,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=False,
        )
        mask = None
        generate_on = None
        if self.method == "masking":
            mask = torch.Tensor(
                [i in genes_to_use for i in tot],
            ).to(device=model.device, dtype=torch.bool)
        elif self.method == "generative":
            generate_on = (
                torch.Tensor([model.genes.index(i) for i in genes_to_impute])
                .to(device=model.device)
                .long()
                .unsqueeze(0)
                .repeat(self.batch_size, 1)
            )
        else:
            raise ValueError("need to be one of generative or masking")

        prevplot = model.doplot
        model.doplot = self.doplot
        model.on_predict_epoch_start()
        model.eval()
        device = model.device.type
        rand = random_str()
        dtype = (
            torch.float16
            if type(model.transformer) is FlashTransformer
            else model.dtype
        )
        save_expr = model.save_expr
        model.save_expr = True
        torch.cuda.empty_cache()
        with torch.no_grad(), torch.autocast(device_type=device, dtype=dtype):
            for batch in tqdm(dataloader):
                gene_pos, expression, depth = (
                    batch["genes"].to(device),
                    batch["x"].to(device),
                    batch["depth"].to(device),
                )
                model._predict(
                    gene_pos,
                    expression,
                    depth,
                    knn_cells=(
                        batch["knn_cells"].to(device)
                        if model.expr_emb_style == "metacell" and self.use_knn
                        else None
                    ),
                    do_generate=self.method == "generative",
                    depth_mult=self.predict_depth_mult,
                    max_size_in_mem=self.save_every,
                    name="impute" + rand + "_",
                    mask=mask,
                    generate_on=generate_on,
                )
        torch.cuda.empty_cache()
        model.log_adata(name="impute" + rand + "_" + str(model.counter))
        try:
            mdir = (
                model.logger.save_dir if model.logger.save_dir is not None else "data"
            )
        except:
            mdir = "data"
        pred_adata = []
        for i in range(model.counter + 1):
            file = (
                mdir
                + "/step_"
                + str(model.global_step)
                + "_"
                + model.name
                + "_impute"
                + rand
                + "_"
                + str(i)
                + "_"
                + str(model.global_rank)
                + ".h5ad"
            )
            pred_adata.append(sc.read_h5ad(file))
            os.remove(file)
        pred_adata = concat(pred_adata)

        model.doplot = prevplot
        model.save_expr = save_expr

        # pred_adata.X = adata.X if random_indices is None else adata.X[random_indices]
        true_imp = pred_adata.X[:, pred_adata.var.index.isin(genes_to_impute)].toarray()

        if true_imp.sum() > 0:
            # we had some gt
            pred_imp = pred_adata.layers["scprint_mu"][
                :, pred_adata.var.index.isin(genes_to_impute)
            ].toarray()
            pred_known = pred_adata.layers["scprint_mu"][
                :, pred_adata.var.index.isin(genes_to_use)
            ].toarray()
            true_known = pred_adata.X[
                :, pred_adata.var.index.isin(genes_to_use)
            ].toarray()

            if self.apply_zero_pred:
                pred_imp = (
                    pred_imp
                    * (
                        1
                        - F.sigmoid(
                            torch.Tensor(
                                pred_adata.layers["scprint_pi"][
                                    :, pred_adata.var.index.isin(genes_to_impute)
                                ].toarray()
                            )
                        )
                    ).numpy()
                )
                pred_known = (
                    pred_known
                    * (
                        1
                        - F.sigmoid(
                            torch.Tensor(
                                pred_adata.layers["scprint_pi"][
                                    :, pred_adata.var.index.isin(genes_to_use)
                                ].toarray()
                            )
                        )
                    ).numpy()
                )
            cell_wise_pred = np.array(
                [
                    spearmanr(pred_imp[i], true_imp[i])[0]
                    for i in range(pred_imp.shape[0])
                ]
            )
            cell_wise_known = np.array(
                [
                    spearmanr(pred_known[i], true_known[i])[0]
                    for i in range(pred_known.shape[0])
                ]
            )
            print(
                {
                    "cell_wise_known": np.mean(cell_wise_known),
                    "cell_wise_pred": np.mean(cell_wise_pred),
                }
            )
            if self.doplot:
                print("depth-wise plot")
                plot_cell_depth_wise_corr_improvement(cell_wise_known, cell_wise_pred)

        return random_indices, pred_adata
